# coding: utf-8

from __future__ import division

__status__ = "Development"

import copy
import inspect
import os
import unittest
from shutil import copyfile

from monty.serialization import loadfn
from monty.tempfile import ScratchDir
from pymatgen.analysis.defects.thermodynamics import DefectPhaseDiagram
from pymatgen.analysis.phase_diagram import PhaseDiagram
from pymatgen.core import Composition, Element
from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.ext.matproj import MPRester
from pymatgen.util.testing import PymatgenTest

from doped.pycdt.core._chemical_potentials import (
    ChemPotAnalyzer,
    MPChemPotAnalyzer,
    UserChemPotAnalyzer,
    UserChemPotInputGenerator,
    get_mp_chempots_from_dpd,
)

file_loc = os.path.abspath(os.path.join(__file__, "..", "..", "..", "..", "test_files"))


class DpdAnalyzerTest(PymatgenTest):
    def setUp(self):
        dpd_path = os.path.split(inspect.getfile(DefectPhaseDiagram))[0]
        pymatgen_test_files = os.path.abspath(os.path.join(dpd_path, "tests"))
        vbm_val = 2.6682
        gap = 1.5
        entries = list(
            loadfn(
                os.path.join(pymatgen_test_files, "GaAs_test_defentries.json")
            ).values()
        )
        for entry in entries:
            entry.parameters.update({"vbm": vbm_val})

        self.dpd = DefectPhaseDiagram(entries, vbm_val, gap)

    def test_get_mp_chempots_from_dpd(self):
        cps = get_mp_chempots_from_dpd(self.dpd)
        self.assertEqual(set(["As-GaAs", "Ga-GaAs"]), set(cps.keys()))
        self.assertEqual(
            [-4.6580705550000001, -3.7317319750000006],
            [cps["As-GaAs"][Element("As")], cps["As-GaAs"][Element("Ga")]],
        )
        self.assertEqual(
            [-5.352569090000001, -3.03723344],
            [cps["Ga-GaAs"][Element("As")], cps["Ga-GaAs"][Element("Ga")]],
        )


class ChemPotAnalyzerTest(PymatgenTest):
    def setUp(self):
        self.CPA = ChemPotAnalyzer()

    def test_get_chempots_from_pda(self):
        with MPRester() as mp:
            bulk_ce = mp.get_entry_by_material_id("mp-2534")
            bulk_species_symbol = [s.symbol for s in bulk_ce.composition.elements]
            entries = mp.get_entries_in_chemsys(bulk_species_symbol)

        # intrinsic chempot test
        pd = PhaseDiagram(entries)
        self.CPA.bulk_ce = bulk_ce
        gaas_cp = self.CPA.get_chempots_from_pd(pd)
        self.assertEqual(set(["As-GaAs", "Ga-GaAs"]), set(gaas_cp.keys()))
        self.assertEqual(
            [-4.6580705550000001, -3.7317319750000006],
            [gaas_cp["As-GaAs"][Element("As")], gaas_cp["As-GaAs"][Element("Ga")]],
        )
        self.assertEqual(
            [-5.352569090000001, -3.03723344],
            [gaas_cp["Ga-GaAs"][Element("As")], gaas_cp["Ga-GaAs"][Element("Ga")]],
        )

    def test_diff_bulk_sub_phases(self):
        fl = ["GaAs", "Sb", "GaSb", "Ga"]
        blk, blknom, subnom = self.CPA.diff_bulk_sub_phases(fl, "Sb")
        self.assertEqual(["Ga", "GaAs"], blk)
        self.assertEqual("Ga-GaAs", blknom)
        self.assertEqual("GaSb-Sb", subnom)


class MPChemPotAnalyzerTest(PymatgenTest):
    def setUp(self):
        self.MPCPA = MPChemPotAnalyzer()

    def test_analyze_GGA_chempots(self):
        """
        Test-cases to cover:
            (0) (stable + mpid given = tested in the CPA.test_get_chempots_from_pda unit test)
            (i) user's computed entry is stable w.r.t MP phase diagram; full_sub_approach = False
            (ii) user's computed entry is stable w.r.t MP phase diagram (and not currently in phase diagram);
                full_sub_approach = False
            (iii) not stable, composition exists in list of stable comps of PD; full_sub_approach = False
            (iv) not stable, composition DOESNT exists in list of stable comps of PD; full_sub_approach = False
            (v) one example of full_sub_approach = True... (just do for case with user's entry being stable)
        """
        with MPRester() as mp:
            bulk_ce = mp.get_entry_by_material_id(
                "mp-2534"
            )  # simulates an entry to play with for this unit test

        # test (i) user's computed entry is stable w.r.t MP phase diagram; full_sub_approach = False
        bce = copy.copy(bulk_ce)
        self.MPCPA = MPChemPotAnalyzer(bulk_ce=bce, sub_species=set(["Sb", "In"]))
        cp_fsaf = self.MPCPA.analyze_GGA_chempots(full_sub_approach=False)
        self.assertEqual(
            set(["Ga-GaAs-GaSb-In", "As-GaAs-InAs-SbAs"]), set(cp_fsaf.keys())
        )
        true_answer = [
            cp_fsaf["Ga-GaAs-GaSb-In"][Element(elt)] for elt in ["Sb", "As", "Ga", "In"]
        ]
        self.assertEqual(
            [-4.44626405, -5.352569090000001, -3.03723344, -2.7518583366666665],
            true_answer,
        )
        # true_answer = [cp_fsaf['InAs-GaSb-In-GaAs'][Element(elt)] for elt in ['Sb', 'As', 'Ga', 'In']]
        # self.assertEqual([-4.243837989999999, -5.15014303, -3.239659500000001, -2.72488125],
        #                        true_answer)
        # true_answer = [cp_fsaf['InAs-GaSb-Sb-GaAs'][Element(elt)] for elt in ['Sb', 'As', 'Ga', 'In']]
        # self.assertEqual([-4.127761275, -5.0340663150000005, -3.3557362150000003, -2.8409579649999994],
        #                        true_answer)
        true_answer = [
            cp_fsaf["As-GaAs-InAs-SbAs"][Element(elt)]
            for elt in ["Sb", "As", "Ga", "In"]
        ]
        self.assertEqual(
            [
                -4.1687393749999995,
                -4.658070555,
                -3.7317319750000006,
                -3.2169537249999998,
            ],
            true_answer,
        )

        # test (v) one example of full_sub_approach = True... (just doing for case with user's entry being stable)
        cp_fsat = self.MPCPA.analyze_GGA_chempots(full_sub_approach=True)
        self.assertEqual(
            set(
                [
                    "Ga-GaAs-GaSb-In",
                    "GaAs-InAs-InSb-Sb",
                    "GaAs-In-InAs-InSb",
                    "As-GaAs-InAs-SbAs",
                    "GaAs-GaSb-In-InSb",
                    "GaAs-InAs-Sb-SbAs",
                    "GaAs-GaSb-InSb-Sb",
                ]
            ),
            set(cp_fsat.keys()),
        )
        true_answer = [
            cp_fsat["Ga-GaAs-GaSb-In"][Element(elt)] for elt in ["Sb", "As", "Ga", "In"]
        ]
        self.assertEqual(
            [-4.44626405, -5.352569090000001, -3.03723344, -2.7518583366666665],
            true_answer,
        )
        true_answer = [
            cp_fsat["GaAs-InAs-InSb-Sb"][Element(elt)]
            for elt in ["Sb", "As", "Ga", "In"]
        ]
        self.assertEqual(
            [-4.127761275, -4.895951335, -3.4938511950000004, -2.9790729449999995],
            true_answer,
        )
        true_answer = [
            cp_fsat["GaAs-In-InAs-InSb"][Element(elt)]
            for elt in ["Sb", "As", "Ga", "In"]
        ]
        self.assertEqual(
            [
                -4.354975883333333,
                -5.123165943333333,
                -3.2666365866666673,
                -2.7518583366666665,
            ],
            true_answer,
        )
        true_answer = [
            cp_fsat["As-GaAs-InAs-SbAs"][Element(elt)]
            for elt in ["Sb", "As", "Ga", "In"]
        ]
        self.assertEqual(
            [
                -4.1687393749999995,
                -4.658070555,
                -3.7317319750000006,
                -3.2169537249999998,
            ],
            true_answer,
        )
        true_answer = [
            cp_fsat["GaAs-GaSb-In-InSb"][Element(elt)]
            for elt in ["Sb", "As", "Ga", "In"]
        ]
        self.assertEqual(
            [
                -4.354975883333333,
                -5.2612809233333335,
                -3.1285216066666672,
                -2.7518583366666665,
            ],
            true_answer,
        )
        true_answer = [
            cp_fsat["GaAs-InAs-Sb-SbAs"][Element(elt)]
            for elt in ["Sb", "As", "Ga", "In"]
        ]
        self.assertEqual(
            [
                -4.127761275,
                -4.6990486549999995,
                -3.6907538750000013,
                -3.1759756250000004,
            ],
            true_answer,
        )
        true_answer = [
            cp_fsat["GaAs-GaSb-InSb-Sb"][Element(elt)]
            for elt in ["Sb", "As", "Ga", "In"]
        ]
        self.assertEqual(
            [
                -4.127761275,
                -5.0340663150000005,
                -3.3557362150000003,
                -2.9790729449999995,
            ],
            true_answer,
        )

        # test (iii) not stable, composition exists in list of stable comps of PD; full_sub_approach = False
        us_bce = ComputedEntry(Composition({"Ga": 1, "As": 1}), -8)
        self.MPCPA = MPChemPotAnalyzer(bulk_ce=us_bce)
        cp_fsaf_us_ce = self.MPCPA.analyze_GGA_chempots(full_sub_approach=False)
        self.assertEqual(set(["As-GaAs", "Ga-GaAs"]), set(cp_fsaf_us_ce.keys()))
        self.assertEqual(
            [-4.6580705550000001, -3.7317319750000006],
            [cp_fsaf_us_ce["As-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
        )
        self.assertEqual(
            [-5.352569090000001, -3.03723344],
            [cp_fsaf_us_ce["Ga-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
        )

        # test (ii) user's computed entry is stable w.r.t MP phase diagram
        #       AND  composition DOESNT exists in list of stable comps of PD; full_sub_approach = False
        s_cdne_bce = ComputedEntry(Composition({"Ga": 2, "As": 3}), -21.5)
        self.MPCPA = MPChemPotAnalyzer(bulk_ce=s_cdne_bce)
        cp_fsaf_s_cdne = self.MPCPA.analyze_GGA_chempots(full_sub_approach=False)
        self.assertEqual(set(["Ga2As3-GaAs", "As-Ga2As3"]), set(cp_fsaf_s_cdne.keys()))
        for tested, answer in zip(
            [cp_fsaf_s_cdne["Ga2As3-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
            [-4.720394939999998, -3.669407590000002],
        ):
            self.assertAlmostEqual(answer, tested)
        self.assertEqual(
            [-4.6580705550000001, -3.7628941674999994],
            [cp_fsaf_s_cdne["As-Ga2As3"][Element(elt)] for elt in ["As", "Ga"]],
        )

        # test (iv) not stable, composition DOESNT exists in list of stable comps of PD; full_sub_approach = False
        #       case a) simple 2D phase diagram
        us_cdne_bce_a = ComputedEntry(Composition({"Ga": 2, "As": 3}), -20.0)
        self.MPCPA = MPChemPotAnalyzer(bulk_ce=us_cdne_bce_a)
        cp_fsaf_us_cdne_a = self.MPCPA.analyze_GGA_chempots(full_sub_approach=False)
        self.assertEqual(set(["As-GaAs"]), set(cp_fsaf_us_cdne_a.keys()))
        self.assertEqual(
            [-4.658070555, -3.7317319750000006],
            [cp_fsaf_us_cdne_a["As-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
        )

        #       case b) larger phase diagram
        us_cdne_bce_b = ComputedEntry(Composition({"Ga": 2, "As": 3, "Sb": 2}), -20.0)
        self.MPCPA = MPChemPotAnalyzer(bulk_ce=us_cdne_bce_b)
        cp_fsaf_us_cdne_b = self.MPCPA.analyze_GGA_chempots(full_sub_approach=False)
        self.assertEqual(set(["GaAs-Sb-SbAs"]), set(cp_fsaf_us_cdne_b.keys()))
        self.assertEqual(
            [-4.127761275, -4.6990486549999995, -3.6907538750000013],
            [
                cp_fsaf_us_cdne_b["GaAs-Sb-SbAs"][Element(elt)]
                for elt in ["Sb", "As", "Ga"]
            ],
        )

    def test_get_chempots_from_composition(self):
        bulk_comp = Composition("Cr2O3")
        self.MPCPA = MPChemPotAnalyzer()  # reinitalize
        cro_cp = self.MPCPA.get_chempots_from_composition(bulk_comp)
        self.assertEqual(set(["Cr2O3-CrO2", "Cr-Cr2O3"]), set(cro_cp.keys()))
        self.assertAlmostEqual(-14.635303979999982, cro_cp["Cr2O3-CrO2"][Element("Cr")])
        self.assertAlmostEqual(-5.51908629500001, cro_cp["Cr2O3-CrO2"][Element("O")])
        self.assertAlmostEqual(-9.63670386, cro_cp["Cr-Cr2O3"][Element("Cr")])
        self.assertAlmostEqual(-8.851486374999999, cro_cp["Cr-Cr2O3"][Element("O")])

    def test_get_mp_entries(self):
        # name mp-id of GaAs system and get mp-entries...
        self.MPCPA = MPChemPotAnalyzer(mpid="mp-2534")
        self.MPCPA.get_mp_entries()
        ents = self.MPCPA.entries
        self.assertEqual(set(["bulk_derived", "subs_set"]), set(ents.keys()))
        self.assertTrue(len(ents["bulk_derived"]))


class UserChemPotAnalyzerTest(PymatgenTest):
    def setUp(self):
        with MPRester() as mp:
            self.bulk_ce = mp.get_entry_by_material_id("mp-2534")
        self.UCPA = UserChemPotAnalyzer(bulk_ce=self.bulk_ce)

    def test_read_phase_diagram_and_chempots(self):
        # set up a local phase diagram object...
        # test non mp case,
        with ScratchDir("."):
            os.mkdir("PhaseDiagram")
            os.mkdir("PhaseDiagram/Ga")
            copyfile(
                os.path.join(file_loc, "vasprun.xml_Ga"), "PhaseDiagram/Ga/vasprun.xml"
            )
            os.mkdir("PhaseDiagram/As")
            copyfile(
                os.path.join(file_loc, "vasprun.xml_As"), "PhaseDiagram/As/vasprun.xml"
            )
            os.mkdir("PhaseDiagram/GaAs")
            copyfile(
                os.path.join(file_loc, "vasprun.xml_GaAs"),
                "PhaseDiagram/GaAs/vasprun.xml",
            )
            cp = self.UCPA.read_phase_diagram_and_chempots(
                full_sub_approach=False, include_mp_entries=False
            )
            self.assertEqual(set(["As-GaAs", "Ga-GaAs"]), set(cp.keys()))
            self.assertEqual(
                [-5.3585191833333328, -4.2880321141666675],
                [cp["As-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
            )
            self.assertEqual(
                [-6.0386246900000007, -3.6079266075],
                [cp["Ga-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
            )

        # followed by an case where MP needs to supplement...
        with ScratchDir("."):
            os.mkdir("PhaseDiagram")
            # NO Ga entry included this time
            os.mkdir("PhaseDiagram/As")
            copyfile(
                os.path.join(file_loc, "vasprun.xml_As"), "PhaseDiagram/As/vasprun.xml"
            )
            os.mkdir("PhaseDiagram/GaAs")
            copyfile(
                os.path.join(file_loc, "vasprun.xml_GaAs"),
                "PhaseDiagram/GaAs/vasprun.xml",
            )
            cp = self.UCPA.read_phase_diagram_and_chempots(
                full_sub_approach=False, include_mp_entries=True
            )
            self.assertEqual(set(["As-GaAs", "Ga-GaAs"]), set(cp.keys()))
            self.assertEqual(
                [-5.3585191833333328, -4.2880321141666675],
                [cp["As-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
            )
            self.assertEqual(
                [-6.609317857500001, -3.03723344],
                [cp["Ga-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
            )


class UserChemPotInputGeneratorTest(PymatgenTest):
    def setUp(self):
        self.UCPIGT = UserChemPotInputGenerator(Composition({"Ga": 1, "As": 1}))

    def test_setup_phase_diagram_calculations(self):
        # test that files get craeted for some file of interest...
        with ScratchDir("."):
            self.UCPIGT.setup_phase_diagram_calculations(full_phase_diagram=True)
            self.assertTrue(os.path.exists("PhaseDiagram/"))
            self.assertTrue(os.path.exists("PhaseDiagram/mp-11_As"))
            self.assertTrue(os.path.exists("PhaseDiagram/mp-11_As/POSCAR"))
            self.assertTrue(os.path.exists("PhaseDiagram/mp-142_Ga"))
            self.assertTrue(os.path.exists("PhaseDiagram/mp-142_Ga/POSCAR"))
            self.assertTrue(os.path.exists("PhaseDiagram/mp-2534_GaAs/"))
            self.assertTrue(os.path.exists("PhaseDiagram/mp-2534_GaAs/POSCAR"))


if __name__ == "__main__":
    unittest.main()



# coding: utf-8

from __future__ import division

__status__ = "Development"

import warnings
from copy import deepcopy


class LDAUCorrection(object):
    """
    Correction method for defect formation energies and
    transition levels that implements the scheme proposed by
    Janotti and Van de Walle. The scheme uses the LDA and LDA+U
    gaps and transition levels along with experimental gap to
    compute the corrections to transition levels and energies
    Reference:
        A. Janotti, C. G. Van de Walle, PRB 76, 165202 (2007)
    Args:
        exp_gap: Experimental gap
        ldau_gap: Computed gap with LDA+U (or GGA+U)
        lda_gap: Computed gap with LDA (or GGA)
    """

    def __init__(self, exp_gap, ldau_gap, lda_gap):
        self.exp_gap = exp_gap
        self.ldau_gap = ldau_gap
        self.lda_gap = lda_gap

    def get_transition_correction(self, ldau_transition, lda_transition):
        """
        Correct the LDA+U transition level using LDA and LDA+U transition
        levels and bandgaps and experimental bandgap
        Args:
            ldau_transition: Transition level computed with LDA+U (or GGA+U)
            lda_transition: Transition computed with LDA (or GGA)
        Returns:
            Correction to the input LDA+U transition level
        """
        diff = (ldau_transition - lda_transition) / (self.ldau_gap - self.lda_gap)
        return diff * (self.exp_gap - self.ldau_gap)

    def get_energy_correction(self, occupancy, ldau_transition, lda_transition):
        """
        Correct the LDA+U defect formation energy using LDA and LDA+U
        transition levels and bandgaps and experimental bandgap
        Args:
            occupancy: Defect level occupancy
            ldau_transition: Transition level computed with LDA+U (or GGA+U)
            lda_transition: Transition computed with LDA (or GGA)
        """
        trans_correction = self.get_transition_correction(
            ldau_transition, lda_transition
        )
        return trans_correction * occupancy


def get_ldau_corrections(
    exp_gap, ldau_gap, lda_gap, ldau_trans, lda_trans, occupancies
):
    """
        NOTE from developers:
            As of 12/15/17 this code does not have unit test
            TODO: benchmark implementation and make unit test
    Compute the corrections to LDA+U computed transition levels and defect
    formation energies using the scheme by Janotti and van de Walle
    Reference:
        A. Janotti, C. G. Van de Walle, PRB 76, 165202 (2007)
    Args:
        exp_gap: Experimental bandgap
        ldau_gap: Computed gap with LDA+U (or GGA+U)
        lda_gap: Computed gap with LDA (or GGA)
        ldau_trans: LDA+U Transition levels between defect charges as dict
            Ex: {'vac_1_Cr': {(0,-1): 0.3, (0,-2): 0.1}}
        lda_trans: LDA Transition levels between defect charges as dict
            Ex: {'vac_1_Cr': {(0,-1): 0.3, (0,-2): 0.1}}
        occupancies: Occupancy of defect levels by electrons as dict
            Ex: {'vac_1_Cr': {0: 0, -2: 2}}
    Returns:
        (transition_level_corrections, energy_corrections)
    """
    # TODO: Use better variable names
    print("occ inside ggau_corr func")
    print(occupancies)
    energy_corrections = {}
    transition_corrections = {}
    corrector = LDAUCorrection(exp_gap, ldau_gap, lda_gap)
    for defect_name in ldau_trans:
        print("def_name", defect_name)
        energy_corrections[defect_name] = {}
        transition_corrections[defect_name] = {}
        ggau_levels = ldau_trans[defect_name]
        gga_levels = lda_trans[defect_name]
        occ = occupancies[defect_name]
        print(occ)
        zero_occ_q = occ["0_occupancy"]
        print(defect_name, zero_occ_q)
        for trans_pair in ggau_levels:
            ggau_transit = ggau_levels[trans_pair]
            search_val = set(list(trans_pair))
            for trans_pair1 in gga_levels:
                match_val = set(list(trans_pair1))
                if match_val == search_val:
                    gga_transit = gga_levels[trans_pair1]
                    break
            q = (set(list(deepcopy(trans_pair))) - set([zero_occ_q])).pop()
            q_occ = occ[q]

            trans_corr = corrector.get_transition_correction(ggau_transit, gga_transit)
            print("trans_corr", defect_name, trans_pair, trans_corr)

            transition_corrections[defect_name][trans_pair] = trans_corr
            new_transit = ggau_transit + trans_corr
            print("new_level", defect_name, trans_pair, new_transit)

            if zero_occ_q in trans_pair:
                enrgy_corr = corrector.get_energy_correction(
                    q_occ, ggau_transit, gga_transit
                )
                energy_corrections[defect_name][q] = enrgy_corr

    return transition_corrections, energy_corrections



# coding: utf-8

from __future__ import division

__status__ = "Development"

import os
import tarfile
import unittest
from shutil import copyfile

from monty.json import MontyDecoder, MontyEncoder
from monty.serialization import dumpfn, loadfn
from monty.tempfile import ScratchDir
from pymatgen import __file__ as initfilep
from pymatgen.core import Element
from pymatgen.core.lattice import Lattice
from pymatgen.core.sites import PeriodicSite
from pymatgen.entries.computed_entries import ComputedStructureEntry
from pymatgen.io.vasp import Locpot
from pymatgen.util.testing import PymatgenTest

from doped.pycdt.core.defects_analyzer import (
    ComputedDefect,
    DefectsAnalyzer,
    freysoldt_correction_from_paths,
    kumagai_correction_from_paths,
)

pmgtestfiles_loc = os.path.join(
    os.path.split(os.path.split(initfilep)[0])[0], "test_files"
)
file_loc = os.path.abspath(os.path.join(__file__, "..", "..", "..", "..", "test_files"))


class FilePathCorrectionsTest(PymatgenTest):
    def test_freysoldt_and_kumagai(self):
        # create scratch directory with files....
        # having to do it all at once to minimize amount of time copying over to Scratch Directory
        with ScratchDir("."):
            # setup with fake Locpot object copied over
            copyfile(
                os.path.join(file_loc, "test_path_files.tar.gz"),
                "./test_path_files.tar.gz",
            )
            tar = tarfile.open("test_path_files.tar.gz")
            tar.extractall()
            tar.close()
            blocpot = Locpot.from_file(os.path.join(file_loc, "bLOCPOT.gz"))
            blocpot.write_file("test_path_files/bulk/LOCPOT")
            dlocpot = Locpot.from_file(os.path.join(file_loc, "dLOCPOT.gz"))
            dlocpot.write_file("test_path_files/sub_1_Sb_on_Ga/charge_2/LOCPOT")

            fcc = freysoldt_correction_from_paths(
                "test_path_files/sub_1_Sb_on_Ga/charge_2/",
                "test_path_files/bulk/",
                18.12,
                2,
                plot=True,
            )
            self.assertEqual(fcc, -1.2435280589593547)
            self.assertTrue(
                os.path.exists(
                    "test_path_files/sub_1_Sb_on_Ga/charge_2/Sub_Sb_on_Ga_mult32_chg_2_axis1_freysoldtplot.pdf"
                )
            )

            kcc = kumagai_correction_from_paths(
                "test_path_files/sub_1_Sb_on_Ga/charge_2/",
                "test_path_files/bulk/",
                18.12,
                2,
                plot=True,
            )
            self.assertEqual(kcc, 0.6387768530616106)
            self.assertTrue(
                os.path.exists(
                    "test_path_files/sub_1_Sb_on_Ga/charge_2/Sub_Sb_on_Ga_mult32_chg_2_kumagaiplot.pdf"
                )
            )


class ComputedDefectTest(PymatgenTest):
    def setUp(self):
        entry_file = os.path.join(file_loc, "vac_cr2o3_struct_entry.json")
        entry = loadfn(entry_file, cls=MontyDecoder)
        lattice = Lattice(
            [
                [9.995004137201189, -2.1469568e-08, 0.0],
                [-4.997501105922451, 8.655927903729987, 0.0],
                [0.0, 0.0, 13.67956098598296],
            ]
        )
        coords = [0.1666665000000016, 0.3333334999999984, 0.014505185117094302]
        site_in_bulk = PeriodicSite("Cr", coords, lattice)
        multiplicity = 12
        supercell_size = [2, 2, 1]
        q = -2
        q_corr = 0.98716
        o_corr = 0.39139821874799996
        name = "vac_1_Cr"
        self.com_def = ComputedDefect(
            entry, site_in_bulk, multiplicity, supercell_size, q, q_corr, o_corr, name
        )

    def test_as_from_dict(self):
        d = self.com_def.as_dict()
        comp_def = ComputedDefect.from_dict(d)
        self.assertIsInstance(comp_def, ComputedDefect)
        with ScratchDir("."):
            dumpfn(self.com_def, "tmp.json", cls=MontyEncoder)
            comp_def = loadfn("tmp.json", cls=MontyDecoder)
            self.assertIsInstance(comp_def, ComputedDefect)


class DefectsAnalyzerTest(PymatgenTest):
    def setUp(self):
        blk_entry_file = os.path.join(file_loc, "Cr2O3_defects.json")
        blk_entry = loadfn(blk_entry_file, cls=MontyDecoder)
        bulk_struct = blk_entry["bulk"]["supercell"]["structure"]
        bulk_energy = -100
        bulk_entry = ComputedStructureEntry(bulk_struct, bulk_energy)
        e_vbm = 0.5
        mu_elts = {Element("Cr"): -10, Element("O"): -5}
        bandgap = 3.0
        self.da = DefectsAnalyzer(bulk_entry, e_vbm, mu_elts, bandgap)

        d1_entry_file = os.path.join(file_loc, "Cr2O3_defects.json")
        d1_entry = loadfn(d1_entry_file, cls=MontyDecoder)
        structure = d1_entry["vacancies"][0]["supercell"]["structure"]
        site_in_bulk = d1_entry["vacancies"][0]["bulk_supercell_site"]
        mult = d1_entry["vacancies"][0]["site_multiplicity"]
        sc_size = d1_entry["vacancies"][0]["supercell"]["size"]
        entry_defect = ComputedStructureEntry(structure, -99)
        self.cd = ComputedDefect(
            entry_defect,
            site_in_bulk,
            multiplicity=mult,
            supercell_size=sc_size,
            charge=2,
            name="vac_1_Cr",
        )

        entry_defect2 = ComputedStructureEntry(structure, -99)
        self.cd2 = ComputedDefect(
            entry_defect2,
            site_in_bulk,
            multiplicity=mult,
            supercell_size=sc_size,
            charge=1,
            name="vac_1_Cr",
        )

    def test_as_from_dict(self):
        d = self.da.as_dict()
        da = DefectsAnalyzer.from_dict(d)
        self.assertIsInstance(da, DefectsAnalyzer)
        with ScratchDir("."):
            dumpfn(self.da, "tmp.json", cls=MontyEncoder)
            da = loadfn("tmp.json", cls=MontyDecoder)
            self.assertIsInstance(da, DefectsAnalyzer)

    def test_add_computed_defect(self):
        self.da.add_computed_defect(self.cd)
        self.assertEqual(len(self.da._defects), 1)
        self.assertEqual(self.da._formation_energies[0], -3.0)

    def test_change_charge_correction(self):
        self.da.add_computed_defect(self.cd)
        self.assertEqual(self.da._defects[0].charge_correction, 0.0)
        self.da.change_charge_correction(0, -1.0)
        self.assertEqual(self.da._defects[0].charge_correction, -1.0)
        self.assertEqual(self.da._formation_energies[0], -4.0)

    def test_change_other_correction(self):
        self.da.add_computed_defect(self.cd)
        self.assertEqual(self.da._defects[0].other_correction, 0.0)
        self.da.change_other_correction(0, -1.5)
        self.assertEqual(self.da._defects[0].other_correction, -1.5)
        self.assertEqual(self.da._formation_energies[0], -4.5)

    def test_get_all_defect_types(self):
        self.da.add_computed_defect(self.cd)
        self.assertArrayEqual(self.da._get_all_defect_types(), ["vac_1_Cr"])

    def test_compute_form_en(self):
        self.da.add_computed_defect(self.cd)
        self.assertEqual(self.da._formation_energies[0], -3.0)

    def test_correct_bg_simple(self):
        self.assertEqual(self.da._e_vbm, 0.5)
        self.assertEqual(self.da._band_gap, 3.0)
        self.da.correct_bg_simple(0.3, 0.5)
        self.assertEqual(self.da._e_vbm, 0.2)
        self.assertEqual(self.da._band_gap, 3.8)
        self.da.add_computed_defect(self.cd)
        self.assertEqual(self.da._formation_energies[0], -3.6)

    def test_get_transition_levels(self):
        self.da.add_computed_defect(self.cd)
        self.da.add_computed_defect(self.cd2)
        self.assertEqual(
            list(self.da.get_transition_levels()["vac_1_Cr"].keys()), [(1, 2)]
        )
        self.assertEqual(self.da.get_transition_levels()["vac_1_Cr"][(1, 2)], -0.5)

    def test_get_form_energy(self):
        self.da.add_computed_defect(self.cd)
        self.assertEqual(self.da._get_form_energy(0.5, 0), -2.0)

    def test_get_formation_energies(self):
        self.da.add_computed_defect(self.cd)
        self.da.add_computed_defect(self.cd2)
        list_fe = self.da.get_formation_energies(ef=0.5)
        self.assertArrayEqual(
            [list_fe[0]["energy"], list_fe[1]["energy"]], [-2.0, -3.0]
        )

    def test_get_defects_concentration(self):
        self.da.add_computed_defect(self.cd)
        self.da.add_computed_defect(self.cd2)
        list_c = self.da.get_defects_concentration(temp=300.0, ef=0.5)
        self.assertArrayEqual(
            [list_c[0]["conc"], list_c[1]["conc"]],
            [2.3075483087087652e62, 1.453493521232979e79],
        )
        list_c = self.da.get_defects_concentration(temp=1000.0, ef=0.5)
        self.assertArrayEqual(
            [list_c[0]["conc"], list_c[1]["conc"]],
            [6.9852762150255027e38, 7.6553010344336244e43],
        )

    def test_get_dos(self):
        dosval = self.da._get_dos(-1.0, 2.0, 3.0, 4.0, -1.4)
        self.assertEqual(dosval, 1.5568745675641716e45)

    def test_get_dos_fd_elec(self):
        elecval = self.da._get_dos_fd_elec(3.1, 2.9, 300.0, 2.0, 3.0, 4.0)
        self.assertEqual(elecval, 9.300000567970405e24)

    def test_get_dos_fd_hole(self):
        holeval = self.da._get_dos_fd_hole(-0.1, 0.2, 300.0, 2.0, 3.0, 4.0)
        self.assertEqual(holeval, 1.9442064601243644e23)

    def test_get_qd(self):
        self.da.add_computed_defect(self.cd)
        self.da.add_computed_defect(self.cd2)
        val = self.da._get_qd(0.5, 300.0)
        self.assertEqual(val, 1.453493521232979e79)

    def test_get_qi(self):
        val = self.da.get_qi(0.1, 300.0, [1.0, 2.0, 3.0], [4.0, 5.0, 6.0])
        self.assertEqual(val, 1.151292510656441e25)

    def test_get_qtot(self):
        self.da.add_computed_defect(self.cd)
        self.da.add_computed_defect(self.cd2)
        val = self.da._get_qtot(0.1, 300.0, [1.0, 2.0, 3.0], [4.0, 5.0, 6.0])
        self.assertEqual(val, 7.6228613357589505e85)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

from __future__ import print_function

__status__ = "Development"

import functools
import os
from copy import deepcopy

import numpy as np
from monty.json import MontyEncoder
from monty.os.path import zpath
from monty.serialization import dumpfn, loadfn
from pymatgen.io.vasp.inputs import Kpoints, Potcar, PotcarSingle
from pymatgen.io.vasp.sets import MPRelaxSet, MPStaticSet, DictSet


def _check_psp_dir():  # Provided by Katarina Brlec, from github.com/SMTG-UCL/surfaxe
    """
    Helper function to check if potcars are set up correctly for use with
    pymatgen, to be compatible across pymatgen versions (breaking changes in v2022)
    """
    potcar = False
    try:
        import pymatgen.settings

        pmg_settings = pymatgen.settings.SETTINGS
        if "PMG_VASP_PSP_DIR" in pmg_settings:
            potcar = True
    except ModuleNotFoundError:
        try:
            import pymatgen

            pmg_settings = pymatgen.SETTINGS
            if "PMG_VASP_PSP_DIR" in pmg_settings:
                potcar = True
        except AttributeError:
            from pymatgen.core import SETTINGS

            pmg_settings = SETTINGS
            if "PMG_VASP_PSP_DIR" in pmg_settings:
                potcar = True
    return potcar


def _import_psp():
    """import pmg settings for PotcarSingleMod"""
    pmg_settings = None
    try:
        import pymatgen.settings

        pmg_settings = pymatgen.settings.SETTINGS
    except ModuleNotFoundError:
        try:
            import pymatgen

            pmg_settings = pymatgen.SETTINGS
        except AttributeError:
            from pymatgen.core import SETTINGS

            pmg_settings = SETTINGS

    if pmg_settings is None:
        raise ValueError("pymatgen settings not found?")
    else:
        return pmg_settings


class PotcarSingleMod(PotcarSingle):
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

    # @staticmethod
    # def from_file(filename):
    #    with smart_open.smart_open(filename, "rt") as f:
    #        return PotcarSingle(f.read())

    @staticmethod
    def from_symbol_and_functional(symbol, functional=None):
        settings = _import_psp()
        if functional is None:
            functional = settings.get("PMG_DEFAULT_FUNCTIONAL", "PBE_54")
        funcdir = PotcarSingle.functional_dir[functional]

        if not os.path.isdir(os.path.join(settings.get("PMG_VASP_PSP_DIR"), funcdir)):
            functional_dir = {
                "LDA_US": "pot",
                "PW91_US": "pot_GGA",
                "LDA": "potpaw",
                "PW91": "potpaw_GGA",
                "LDA_52": "potpaw_LDA.52",
                "LDA_54": "potpaw_LDA.54",
                "PBE": "potpaw_PBE",
                "PBE_52": "potpaw_PBE.52",
                "PBE_54": "potpaw_PBE.54",
            }
            funcdir = functional_dir[functional]

        d = settings.get("PMG_VASP_PSP_DIR")
        if d is None:
            raise ValueError(
                "No POTCAR directory found. Please set "
                "the VASP_PSP_DIR environment variable"
            )

        paths_to_try = [
            os.path.join(d, funcdir, "POTCAR.{}".format(symbol)),
            os.path.join(d, funcdir, symbol, "POTCAR.Z"),
            os.path.join(d, funcdir, symbol, "POTCAR"),
        ]
        for p in paths_to_try:
            p = os.path.expanduser(p)
            p = zpath(p)
            if os.path.exists(p):
                return PotcarSingleMod.from_file(p)
        raise IOError(
            "You do not have the right POTCAR with functional "
            + "{} and label {} in your VASP_PSP_DIR".format(functional, symbol)
        )


class PotcarMod(Potcar):
    def __init__(self, **kwargs):
        super(self.__class__, self).__init__(**kwargs)

    def set_symbols(self, symbols, functional=None, sym_potcar_map=None):
        """
        Initialize the POTCAR from a set of symbols. Currently, the POTCARs can
        be fetched from a location specified in .pmgrc.yaml. Use pmg config
        to add this setting.

        Args:
            symbols ([str]): A list of element symbols
            functional (str): The functional to use. If None, the setting
                PMG_DEFAULT_FUNCTIONAL in .pmgrc.yaml is used, or if this is
                not set, it will default to PBE_54.
            sym_potcar_map (dict): A map of symbol:raw POTCAR string. If
                sym_potcar_map is specified, POTCARs will be generated from
                the given map data rather than the config file location.
        """
        del self[:]
        if sym_potcar_map:
            for el in symbols:
                self.append(PotcarSingleMod(sym_potcar_map[el]))
        else:
            for el in symbols:
                p = PotcarSingleMod.from_symbol_and_functional(el, functional)
                self.append(p)


class DefectRelaxSet(DictSet):
    """
    pymatgen DictSet for VASP Defect Relaxation Calculations
    Additional Args:
        charge: Charge of the defect structure
    """

    def __init__(self, structure, **kwargs):
        charge = kwargs.pop("charge", 0)
        config_dict = kwargs.pop("config_dict", {})
        if config_dict:
            super(self.__class__, self).__init__(structure, config_dict=config_dict, **kwargs)
        else:
            mp_set = MPRelaxSet(structure, **kwargs)
            super(self.__class__, self).__init__(structure, config_dict=mp_set.CONFIG, **kwargs)

        self.charge = charge

    @property
    def incar(self):
        inc = super(self.__class__, self).incar
        try:
            inc["NELECT"] = self.nelect - self.charge
            if inc["NELECT"] % 2 != 0:  # odd number of electrons
                inc["NUPDOWN"] = 1
        except Exception:
            print("NELECT and NUPDOWN flags are not set due to non-availability of POTCARs")

        return inc

    @property
    def potcar(self):
        """
        Potcar object.
        """
        return PotcarMod(symbols=self.potcar_symbols, functional=self.potcar_functional)

    @property
    def all_input(self):
        """
        Returns all input files as a dict of {filename: vasp object}

        Returns:
            dict of {filename: object}, e.g., {'INCAR': Incar object, ...}
        """
        try:
            return super(DefectRelaxSet, self).all_input
        except:  # Expecting the error to be POTCAR related, its ignored
            kpoints = self.kpoints
            incar = self.incar
            if np.product(kpoints.kpts) < 4 and incar.get("ISMEAR", 0) == -5:
                incar["ISMEAR"] = 0

            return {"INCAR": incar, "KPOINTS": kpoints, "POSCAR": self.poscar}

    @property
    def structure(self):
        """
        :return: Structure
        """
        return self._structure


class DefectStaticSet(MPStaticSet):
    """
    Extension to MPStaticSet which modifies some parameters appropriate
    for bulk supercell calculation
    """

    def __init__(self, structure, **kwargs):
        super(self.__class__, self).__init__(structure, **kwargs)

    @property
    def potcar(self):
        """
        Potcar object.
        """
        return PotcarMod(symbols=self.potcar_symbols, functional=self.potcar_functional)

    @property
    def all_input(self):
        """
        Returns all input files as a dict of {filename: vasp object}

        Returns:
            dict of {filename: object}, e.g., {'INCAR': Incar object, ...}
        """
        try:
            return super(DefectStaticSet, self).all_input
        except:
            kpoints = self.kpoints
            incar = self.incar
            if np.product(kpoints.kpts) < 4 and incar.get("ISMEAR", 0) == -5:
                incar["ISMEAR"] = 0

            return {"INCAR": incar, "KPOINTS": kpoints, "POSCAR": self.poscar}


class DielectricSet(MPStaticSet):
    """
    Extension to MPStaticSet which modifies some parameters appropriate
    for bulk supercell calculation
    """

    def __init__(self, structure, **kwargs):
        super(self.__class__, self).__init__(structure, lepsilon=True, **kwargs)

    @property
    def potcar(self):
        """
        Potcar object.
        """
        return PotcarMod(symbols=self.potcar_symbols, functional=self.potcar_functional)

    @property
    def all_input(self):
        """
        Returns all input files as a dict of {filename: vasp object}

        Returns:
            dict of {filename: object}, e.g., {'INCAR': Incar object, ...}
        """
        try:
            return super(self.__class__, self).all_input
        except:
            kpoints = self.kpoints
            incar = self.incar
            if np.product(kpoints.kpts) < 4 and incar.get("ISMEAR", 0) == -5:
                incar["ISMEAR"] = 0

            return {"INCAR": incar, "KPOINTS": kpoints, "POSCAR": self.poscar}


def write_additional_files(path, trans_dict=None, incar={}, kpoints=None, hse=False):
    """
        NOTE from developers:
            This code is primarily used for dumping extra files
            for users desiring to do HSE
            will not be used in command line code or maintained (with unit tests etc.)
            going forward (as of 12/15/2017).
            However, we are keeping function here to allow for
            current users to make use of it...

    Write the additional files based on user settings
    """
    if kpoints:
        kpoints.write_file(os.path.join(path, "KPOINTS"))

    if trans_dict:
        dumpfn(trans_dict, os.path.join(path, "transformation.json"), cls=MontyEncoder)

    if hse:
        if not incar:
            raise ValueError("Incar settings need to passed to write incar")

        incar.update({"LWAVE": True})
        incar.write_file(os.path.join(path, "INCAR.gga"))
        incar.update(
            {
                "LHFCALC": True,
                "ALGO": "All",
                "HFSCREEN": 0.2,
                "PRECFOCK": "Fast",
                "NKRED": 2,
            }
        )
        incar.write_file(os.path.join(path, "INCAR.hse1"))
        del incar["PRECFOCK"]
        del incar["NKRED"]
        incar.write_file(os.path.join(path, "INCAR.hse2"))


def make_vasp_defect_files(defects, path_base, user_settings={}, hse=False):
    """
    Generates VASP files for defect computations
    Args:
        defect_structs:
            the defects data as a dictionnary. Ideally this is generated
            from core.defectsmaker.ChargedDefectsStructures.
        path_base:
            where we write the files
        user_settings:
            Settings in dict format to override the defaults used in
            generating vasp files. The format of the dictionary is
            {'defects:{'INCAR':{...},'KPOINTS':{...},
             'bulk':{'INCAR':{...},'KPOINTS':{...}}
        hse:
            hse run or not
    """
    bulk_sys = defects["bulk"]["supercell"]
    comb_defs = functools.reduce(
        lambda x, y: x + y, [defects[key] for key in defects if key != "bulk"]
    )

    # User setting dicts
    user_settings = deepcopy(user_settings)
    user_incar = user_settings.pop("INCAR", {})
    user_incar_blk_tmp = user_incar.pop("bulk", {})
    user_incar_blk_def = user_incar.pop("defects", {})
    user_incar.pop("dielectric", {})
    user_incar_blk = deepcopy(user_incar)
    user_incar_def = deepcopy(user_incar)
    user_incar_blk.update(user_incar_blk_tmp)
    user_incar_def.update(user_incar_blk_def)
    user_kpoints = user_settings.pop("KPOINTS", {})
    potcar_settings = user_settings.pop("POTCAR", {})
    potcar_functional = potcar_settings.pop("functional", "PBE_54")

    for defect in comb_defs:
        for charge in defect["charges"]:
            s = defect["supercell"]
            dict_transf = {
                "defect_type": defect["name"],
                "defect_site": defect["unique_site"],
                "defect_supercell_site": defect["bulk_supercell_site"],
                "defect_multiplicity": defect["site_multiplicity"],
                "charge": charge,
                "supercell": s["size"],
            }
            if "substitution_specie" in defect:
                dict_transf["substitution_specie"] = defect["substitution_specie"]

            defect_relax_set = DefectRelaxSet(
                s["structure"],
                user_incar_settings=user_incar_def,
                user_potcar_settings=potcar_settings,
                potcar_functional=potcar_functional,
                charge=charge,
            )

            path = os.path.join(path_base, defect["name"], "charge_" + str(charge))
            try:
                potcar = defect_relax_set.potcar
            except:
                potcar = None

            if potcar or not charge:
                defect_relax_set.write_input(path)
                incar = defect_relax_set.incar if hse else {}
                kpoints = Kpoints.from_dict(user_kpoints) if user_kpoints else None

                write_additional_files(
                    path, dict_transf, incar=incar, kpoints=kpoints, hse=hse
                )
            else:
                os.makedirs(path)
                with open(os.path.join(path, "readme.txt"), "w") as fp:
                    print(
                        "Vasp input files not generated for charged defects "
                        "due to unavailability of POTCAR. "
                        "If charged defects desired, please supply POTCAR file "
                        "path to .pmgrc.yaml file.",
                        end="",
                        file=fp,
                    )

    # Generate bulk supercell inputs
    s = bulk_sys
    dict_transf = {"defect_type": "bulk", "supercell": s["size"]}

    # potcar_functional = user_potcar.get('functional', 'PBE_54')
    blk_static_set = DefectStaticSet(
        s["structure"],
        user_incar_settings=user_incar_blk,
        user_potcar_settings=potcar_settings,
        potcar_functional=potcar_functional,
    )
    path = os.path.join(path_base, "bulk")
    blk_static_set.write_input(path)

    incar = blk_static_set.incar if hse else {}
    kpoints = Kpoints.from_dict(user_kpoints) if user_kpoints else None

    write_additional_files(path, dict_transf, incar=incar, kpoints=kpoints, hse=hse)


def make_vasp_defect_files_dos(
    defects, path_base, user_settings={}, hse=False, dos_limits=(-1, 7)
):
    """
        NOTE from developers:
            This code will not be used in command line code or
            maintained going forward (as of 12/15/2017).
            However, we are keeping function here to allow for
            current users to make use of it...

    Generates VASP files for defect computations which include dos
    generation. Useful when the user don't want to use MPWorks for
    dos calculations.
    Args:
        defects:
            the defects data as a dictionnary. Ideally this is generated
            from core.defectsmaker.ChargedDefectsStructures.
        path_base:
            where we write the files
        user_settings:
            Settings in dict format to override the defaults used in
            generating vasp files. The format of the dictionary is
            {'defects:{'INCAR':{...},'KPOINTS':{...},
             'bulk':{'INCAR':{...},'KPOINTS':{...}}
        hse:
            hse run or not
        dos_limits:
            Lower and upper limits for dos plot as a tuple. The default
            (-1,7) should work for most of the cases.
    """
    bulk_sys = defects["bulk"]["supercell"]
    comb_defs = functools.reduce(
        lambda x, y: x + y, [defects[key] for key in defects if key != "bulk"]
    )

    for defect in comb_defs:
        for charge in defect["charges"]:
            s = defect["supercell"]
            dict_transf = {
                "defect_type": defect["name"],
                "defect_site": defect["unique_site"],
                "defect_supercell_site": defect["bulk_supercell_site"],
                "charge": charge,
                "supercell": s["size"],
            }
            if "substitution_specie" in defect:
                dict_transf["substitution_specie"] = defect["substitution_specie"]

            mp_relax_set = MPRelaxSet(s["structure"])
            incar = mp_relax_set.incar

            incar.update(
                {
                    "IBRION": 2,
                    "ISIF": 2,
                    "ISPIN": 2,
                    "LWAVE": False,
                    "EDIFF": 1e-5,
                    "EDIFFG": -1e-2,
                    "ISMEAR": 0,
                    "SIGMA": 0.05,
                    "LVTOT": True,
                    "LVHAR": True,
                    "LORBIT": 14,
                    "ALGO": "Fast",
                    "ISYM": 0,
                }
            )
            if user_settings:
                if "INCAR" in user_settings.get("defects", {}):
                    incar.update(user_settings["defects"]["INCAR"])

            comp = s["structure"].composition
            sum_elec = 0
            elts = set()
            for p in mp_relax_set.potcar:
                if p.element not in elts:
                    sum_elec += comp.as_dict()[p.element] * p.nelectrons
                    elts.add(p.element)

            incar["NELECT"] = sum_elec - charge

            kpoint = mp_relax_set.kpoints.monkhorst_automatic()
            path = os.path.join(path_base, defect["name"], "charge_" + str(charge))
            try:
                os.makedirs(path)
            except:
                pass
            if hse:
                incar.update({"LWAVE": True})
                incar.write_file(os.path.join(path, "INCAR.relax.gga"))
                incar.update(
                    {
                        "LHFCALC": True,
                        "ALGO": "All",
                        "HFSCREEN": 0.208,
                        "PRECFOCK": "Fast",
                        "NKRED": 2,
                    }
                )
                incar.write_file(os.path.join(path, "INCAR.relax.hse1"))
                del incar["PRECFOCK"]
                del incar["NKRED"]
                incar.write_file(os.path.join(path, "INCAR.relax.hse2"))
            else:
                incar.write_file(os.path.join(path, "INCAR.relax"))
            kpoint.write_file(os.path.join(path, "KPOINTS"))

            mp_relax_set.poscar.write_file(os.path.join(path, "POSCAR.orig"))
            mp_relax_set.potcar.write_file(os.path.join(path, "POTCAR"))
            dumpfn(
                dict_transf, os.path.join(path, "transformation.json"), cls=MontyEncoder
            )

            # Write addition incar files for dos plots of defect levels
            del incar["NSW"]
            del incar["ISIF"]
            del incar["EDIFFG"]
            del incar["LVHAR"]
            del incar["LVTOT"]
            incar["IBRION"] = -1
            incar["ICHARG"] = 1
            incar["EDIFF"] = 1e-6
            # High acc not required for chgcar
            incar.update({"PRECFOCK": "Fast", "NKRED": 2})
            incar.write_file(os.path.join(path, "INCAR.static"))
            del incar["PRECFOCK"]
            del incar["NKRED"]
            incar["ICHARG"] = 11
            incar["NEDOS"] = int((dos_limits[1] - dos_limits[0]) / 0.006)
            incar["EMIN"] = dos_limits[0]
            incar["EMAX"] = dos_limits[1]
            incar.write_file(os.path.join(path, "INCAR.dos"))

    # Generate bulk supercell inputs
    s = bulk_sys
    dict_transf = {"defect_type": "bulk", "supercell": s["size"]}

    mp_relax_set = MPRelaxSet(s["structure"])
    incar = mp_relax_set.incar

    incar.update(
        {
            "IBRION": -1,
            "NSW": 0,
            "ISPIN": 2,
            "LWAVE": False,
            "EDIFF": 1e-5,
            "ISMEAR": 0,
            "SIGMA": 0.05,
            "LVTOT": True,
            "LVHAR": True,
            "ALGO": "Fast",
            "ISYM": 0,
        }
    )
    if user_settings:
        if "INCAR" in user_settings.get("bulk", {}):
            incar.update(user_settings["bulk"]["INCAR"])

    kpoint = mp_relax_set.kpoints.monkhorst_automatic()
    path = os.path.join(path_base, "bulk")
    try:
        os.makedirs(path)
    except:
        pass
    if hse:
        incar.update({"LWAVE": True})
        incar.write_file(os.path.join(path, "INCAR.gga"))
        incar.update(
            {
                "LHFCALC": True,
                "ALGO": "All",
                "HFSCREEN": 0.2,  # "AEXX": 0.45,
                "PRECFOCK": "Fast",
                "NKRED": 2,
            }
        )
        incar.write_file(os.path.join(path, "INCAR.hse1"))
        del incar["PRECFOCK"]
        del incar["NKRED"]
        incar.write_file(os.path.join(path, "INCAR.hse2"))
    else:
        incar.write_file(os.path.join(path, "INCAR"))
    kpoint.write_file(os.path.join(path, "KPOINTS"))
    mp_relax_set.poscar.write_file(os.path.join(path, "POSCAR"))
    mp_relax_set.potcar.write_file(os.path.join(path, "POTCAR"))
    dumpfn(dict_transf, os.path.join(path, "transformation.json"), cls=MontyEncoder)


def make_vasp_dielectric_files(struct, path=None, user_settings={}, hse=False):
    """
    Generates VASP files for dielectric constant computations
    Args:
        struct:
            unitcell in pymatgen structure format
        user_settings:
            Settings in dict format to override the defaults used in
            generating vasp files. The format of the dictionary is
            {'INCAR':{...}, 'KPOINTS':{...}}
        hse:
            hse run or not
    """

    # Generate vasp inputs for dielectric constant
    user_settings = deepcopy(user_settings)
    user_incar = user_settings.pop("INCAR", {})
    user_incar.pop("bulk", {})
    user_incar.pop("defects", {})
    user_incar_diel = user_incar.pop("dielectric", {})
    user_incar.update(user_incar_diel)
    user_kpoints = user_settings.pop("KPOINTS", {})
    grid_density = user_kpoints.get("grid_density", 1000)
    potcar_settings = user_settings.pop("POTCAR", {})
    potcar_functional = potcar_settings.pop("functional", "PBE_54")
    dielectric_set = DielectricSet(
        struct,
        user_incar_settings=user_incar,
        user_potcar_settings=potcar_settings,
        potcar_functional=potcar_functional,
    )

    if not path:
        path_base = struct.composition.reduced_formula
        path = os.path.join(path_base, "dielectric")
    dielectric_set.write_input(path)

    kpoints = Kpoints.automatic_density(struct, grid_density, force_gamma=True)
    incar = dielectric_set.incar if hse else {}
    write_additional_files(path, incar=incar, kpoints=kpoints, hse=hse)



"""
This module contains calls to finite size supercell charge corrections as
implemented in pymatgen.analysis.defects.core

The methods implemented are
1) Freysoldt correction for isotropic systems. Includes:
       a) PC energy
       b) potential alignment by planar averaging.
2) Extended Freysoldt or Kumagai correction for anistropic systems. Includes:
       a) anisotropic PC energy
       b) potential alignment by atomic site averaging outside Wigner Seitz radius
3) Sxdefectalign wrapper - python interface to using robust C++ code written by
   Freysoldt et. al (in principle the results are identical to the output of our own
   Freysoldt python code)

If you use the corrections implemented in this module, cite
   Freysoldt, Neugebauer, and Van de Walle,
    Phys. Status Solidi B. 248, 1067-1076 (2011) for isotropic correction
   Kumagai and Oba, Phys. Rev. B. 89, 195205 (2014) for anisotropic correction
   in addition to the PyCDT paper
"""


import copy
import collections

import numpy as np
from pymatgen.analysis.defects.corrections import FreysoldtCorrection, KumagaiCorrection
from monty.json import MontyDecoder

from doped.pycdt.corrections.sxdefect_correction import SxdefectalignWrapper as SXD


def _monty_decode_nested_dicts(d):
    """
    Recursively find any dictionaries in defect_entry.parameters, which may be nested in dicts or in
    lists of dicts and decode them:
    """
    for key, value in d.items():
        if isinstance(value, dict) and not any(
            k in value for k in ["@module", "@class"]
        ):
            _monty_decode_nested_dicts(value)
        elif isinstance(value, list):
            if all(isinstance(i, dict) for i in value) and all(
                k in i for k in ["@module", "@class"] for i in value
            ):
                try:
                    d[key] = [MontyDecoder().process_decoded(i) for i in value]
                except Exception as exc:
                    print(f"Failed to decode {key} with error {exc}")
                    pass

        if isinstance(value, dict) and all(k in value for k in ["@module", "@class"]):
            try:
                d[key] = MontyDecoder().process_decoded(value)
            except Exception as exc:
                print(f"Failed to decode {key} with error {exc}")
                pass


def get_correction_freysoldt(
    defect_entry, epsilon, plot: bool = False, filename=None, partflag="All", axis=None
):
    """
    Function to compute the isotropic freysoldt correction for each defect.
    If this correction is used, please reference Freysoldt's original paper.
    doi: 10.1103/PhysRevLett.102.016402
    Args:
        defect_entry: DefectEntry object with the following
            keys stored in defect.parameters:
                required:
                    axis_grid (3 x NGX where NGX is the length of the NGX grid
                    in the x,y and z axis directions. Same length as planar
                    average lists):
                        A list of 3 numpy arrays which contain the cartesian axis
                        values (in angstroms) that correspond to each planar avg
                        potential supplied.

                    bulk_planar_averages (3 x NGX where NGX is the length of
                    the NGX grid in the x,y and z axis directions.):
                        A list of 3 numpy arrays which contain the planar averaged
                        electrostatic potential for the bulk supercell.

                    defect_planar_averages (3 x NGX where NGX is the length of
                    the NGX grid in the x,y and z axis directions.):
                        A list of 3 numpy arrays which contain the planar averaged
                        electrostatic potential for the defective supercell.

                    bulk_sc_structure (Structure) bulk structure corresponding to
                        defect supercell structure (uses Lattice for charge correction)

                    defect_frac_sc_coords (3 x 1 array) Fracitional co-ordinates of
                        defect location in supercell structure
                optional:
                    'encut' : energy cutoff desired for Freysoldt correction
                    'madetol' : madelung tolerance for Freysoldt correction
                    'q_model' : Charge Model for Freysoldt correction
                    'q_model' : Charge Model for Freysoldt correction
        epsilon (float or 3x3 matrix): Dielectric constant for the structure
        plot (bool): decides whether to plot electrostatic potential plots or not...
        filename (str): if None, plots are not saved, if a string,
            then the plot will be saved as 'filename_{axis}.pdf'
        partflag: four options for correction output:
               'pc' for just point charge correction, or
               'potalign' for just potalign correction, or
               'All' for both (added together), or
               'AllSplit' for individual parts split up (form is [PC, potterm, full])
        axis (int or None): if integer, then freysoldt correction is performed on the single axis.
            If it is None, then averaging of the corrections for the three axes is used for the correction.

    Returns Correction
    """
    # ensure parameters are decoded in case defect_dict was reloaded from json
    _monty_decode_nested_dicts(defect_entry.parameters)

    if partflag not in ["All", "AllSplit", "pc", "potalign"]:
        print(
            '{} is incorrect potalign type. Must be "All", "AllSplit", "pc", or '
            '"potalign".'.format(partflag)
        )
        return

    q_model = defect_entry.parameters.get("q_model", None)
    encut = defect_entry.parameters.get("encut", 520)
    madetol = defect_entry.parameters.get("madetol", 0.0001)

    if not defect_entry.charge:
        print("Charge is zero so charge correction is zero.")
        return 0.0

    template_defect = copy.deepcopy(defect_entry)
    corr_class = FreysoldtCorrection(
        epsilon, q_model=q_model, energy_cutoff=encut, madetol=madetol, axis=axis
    )
    f_corr_summ = corr_class.get_correction(template_defect)

    if plot:
        if axis is None:
            ax_list = [
                [k, "axis" + str(k)]
                for k in corr_class.metadata["pot_plot_data"].keys()
            ]
        else:
            ax_list = [[axis, "axis" + str(axis + 1)]]

        for ax_key, ax_title in ax_list:
            p = corr_class.plot(ax_key, title=ax_title, saved=False)
            if filename:
                p.savefig(filename + "_" + ax_title + ".pdf", bbox_inches="tight")

    if partflag in ["AllSplit", "All"]:
        freyval = np.sum(list(f_corr_summ.values()))
    elif partflag == "pc":
        freyval = f_corr_summ["freysoldt_electrostatic"]
    elif partflag == "potalign":
        freyval = f_corr_summ["freysoldt_potential_alignment"]

    print(f"Final Freysoldt correction is {freyval:.3f} eV")

    if partflag == "AllSplit":
        freyval = [
            f_corr_summ["freysoldt_electrostatic"],
            f_corr_summ["freysoldt_potential_alignment"],
            freyval,
        ]

    return freyval


def get_correction_kumagai(defect_entry, epsilon, title=None, partflag="All"):
    """
    Function to compute the Kumagai correction for each defect (modified freysoldt for anisotropic dielectric).
    NOTE that bulk_init class must be pre-instantiated to use this function
    Args:
        defect_entry: DefectEntry object with the following
            keys stored in defect.parameters:
                required:
                    bulk_atomic_site_averages (list):  list of bulk structure"s atomic site averaged ESPs * charge,
                        in same order as indices of bulk structure
                        note this is list given by VASP's OUTCAR (so it is multiplied by a test charge of -1)

                    defect_atomic_site_averages (list):  list of defect structure"s atomic site averaged ESPs * charge,
                        in same order as indices of defect structure
                        note this is list given by VASP's OUTCAR (so it is multiplied by a test charge of -1)

                    site_matching_indices (list):  list of corresponding site index values for
                        bulk and defect site structures EXCLUDING the defect site itself
                        (ex. [[bulk structure site index, defect structure"s corresponding site index], ... ]

                    initial_defect_structure (Structure): Pymatgen Structure object representing un-relaxed defect structure

                    defect_frac_sc_coords (array): Defect Position in fractional coordinates of the supercell
                        given in bulk_structure
                optional:
                    gamma (float): Ewald parameter, Default is to determine it based on convergence of
                        brute summation tolerance
                    sampling_radius (float):r adius (in Angstrom) which sites must be outside of to be included
                        in the correction. Publication by Kumagai advises to use Wigner-Seitz radius of
                        defect supercell, so this is default value.
        epsilon (float or 3x3 matrix): Dielectric constant for the structure
        title: decides whether to plot electrostatic potential plots or not...
            if None, no plot is printed, if a string,
            then the plot will be saved using the string
        partflag: four options for correction output:
               'pc' for just point charge correction, or
               'potalign' for just potalign correction, or
               'All' for both (added together), or
               'AllSplit' for individual parts split up (form is [PC, potterm, full])
    """
    # ensure parameters are decoded in case defect_dict was reloaded from json
    _monty_decode_nested_dicts(defect_entry.parameters)

    if partflag not in ["All", "AllSplit", "pc", "potalign"]:
        print(
            '{} is incorrect potalign type. Must be "All", "AllSplit", "pc", or '
            '"potalign".'.format(partflag)
        )
        return

    sampling_radius = defect_entry.parameters.get("sampling_radius", None)
    gamma = defect_entry.parameters.get("gamma", None)

    if not defect_entry.charge:
        print("Charge is zero so charge correction is zero.")
        return 0.0

    template_defect = copy.deepcopy(defect_entry)
    corr_class = KumagaiCorrection(
        epsilon, sampling_radius=sampling_radius, gamma=gamma
    )
    k_corr_summ = corr_class.get_correction(template_defect)

    if title:
        p = corr_class.plot(title="Kumagai", saved=False)
        p.savefig(title + "_kumagaiplot.pdf", bbox_inches="tight")

    if partflag in ["AllSplit", "All"]:
        kumagai_val = np.sum(list(k_corr_summ.values()))
    elif partflag == "pc":
        kumagai_val = k_corr_summ["kumagai_electrostatic"]
    elif partflag == "potalign":
        kumagai_val = k_corr_summ["kumagai_potential_alignment"]

    print(f"\nFinal Kumagai correction is {kumagai_val:.3f} eV")

    if partflag == "AllSplit":
        kumagai_val = [
            k_corr_summ["kumagai_electrostatic"],
            k_corr_summ["kumagai_potential_alignment"],
            kumagai_val,
        ]
    return kumagai_val


def get_correction_sxdefect(
    path_def,
    path_blk,
    epsilon,
    pos,
    charge,
    title=None,
    lengths=None,
    partflag="All",
    encut=520,
):
    """
        NOTE FROM DEVELOPERS:
        This is not unit tested and will not be maintained past 12/15/17.
        Code remaining here to allow for existing users to keep using it.

    Args:
        lengths: for length conversion (makes calculation faster)
        pos: specify position for sxdefectalign code
        axiscalcs: Specifies axes to average over (zero-defined)
        partflag: four options
            'pc' for just point charge correction, or
           'potalign' for just potalign correction, or
           'All' for both, or
           'AllSplit' for individual parts split up (form [PC,potterm,full])
    """
    # ensure parameters are decoded in case defect_dict was reloaded from json
    _monty_decode_nested_dicts(defect_entry.parameters)

    if partflag in ["All", "AllSplit"]:
        nomtype = "full correction"
    elif partflag == "pc":
        nomtype = "point charge correction"
    elif partflag == "potalign":
        nomtype = "potential alignment correction"
    else:
        print(
            partflag,
            ' is incorrect potalign type. Must be "All","AllSplit", "pc", or "potalign".',
        )
        return

    s = SXD(path_blk, path_def, charge, epsilon, pos, encut, lengths=lengths)

    if title:
        print_flag = "plotfull"
    else:
        print_flag = "none"
    sxvals = s.run_correction(print_pot_flag=print_flag, partflag=partflag)

    print("\n Final Sxdefectalign ", nomtype, " correction value is ", sxvals)

    return sxvals



import contextlib
import copy
import os
import warnings
from pathlib import Path, PurePath

import pandas as pd
from monty.serialization import loadfn
from pymatgen.analysis.phase_diagram import PDEntry, PhaseDiagram
from pymatgen.core import Composition, Element, Structure
from pymatgen.entries.computed_entries import ComputedStructureEntry
from pymatgen.ext.matproj import MPRester
from pymatgen.io.vasp.inputs import Kpoints, UnknownPotcarWarning
from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.io.vasp.sets import BadInputSetWarning, DictSet

from doped.pycdt.utils.parse_calculations import get_vasprun

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
default_potcar_dict = loadfn(os.path.join(MODULE_DIR, "PotcarSet.yaml"))

# globally ignore:
warnings.filterwarnings("ignore", category=BadInputSetWarning)
warnings.filterwarnings("ignore", category=UnknownPotcarWarning)
warnings.filterwarnings("ignore", message="No POTCAR file with matching TITEL fields")
warnings.filterwarnings(
    "ignore", message="POTCAR data with symbol"
)  # Ignore POTCAR warnings because Pymatgen incorrectly detecting POTCAR types
warnings.filterwarnings(
    "ignore", message="You are using the legacy MPRester"
)  # currently rely on this so shouldn't show warning
warnings.filterwarnings("ignore", message="Ignoring unknown variable type")


# TODO: Check default error when user attempts `CompetingPhases()` with no API key setup; if not
#  sufficiently informative, add try except catch to give more informative error message for this.
# TODO: Need to recheck all functionality from old `_chemical_potentials.py` is now present here.


def make_molecule_in_a_box(element):
    # (but do try to fix it so that the nupdown is the same as magnetisation
    # so that it makes that assignment easier later on when making files)
    # the bond distances are taken from various sources and *not* thoroughly vetted
    lattice = [[30, 0, 0], [0, 30, 0], [0, 0, 30]]
    all_structures = {
        "O2": {
            "structure": Structure(
                lattice=lattice,
                species=["O", "O"],
                coords=[[15, 15, 15], [15, 15, 16.22]],
                coords_are_cartesian=True,
            ),
            "formula": "O2",
            "total_magnetization": 2,
        },
        "N2": {
            "structure": Structure(
                lattice=lattice,
                species=["N", "N"],
                coords=[[15, 15, 15], [15, 15, 16.09]],
                coords_are_cartesian=True,
            ),
            "formula": "N2",
            "total_magnetization": 0,
        },
        "H2": {
            "structure": Structure(
                lattice=lattice,
                species=["H", "H"],
                coords=[[15, 15, 15], [15, 15, 15.74]],
                coords_are_cartesian=True,
            ),
            "formula": "H2",
            "total_magnetization": 0,
        },
        "F2": {
            "structure": Structure(
                lattice=lattice,
                species=["F", "F"],
                coords=[[15, 15, 15], [15, 15, 16.44]],
                coords_are_cartesian=True,
            ),
            "formula": "F2",
            "total_magnetization": 0,
        },
        "Cl2": {
            "structure": Structure(
                lattice=lattice,
                species=["Cl", "Cl"],
                coords=[[15, 15, 15], [15, 15, 16.99]],
                coords_are_cartesian=True,
            ),
            "formula": "Cl2",
            "total_magnetization": 0,
        },
    }

    if element in all_structures:
        structure = all_structures[element]["structure"]
        formula = all_structures[element]["formula"]
        total_magnetization = all_structures[element]["total_magnetization"]

    return structure, formula, total_magnetization


def _make_molecular_entry(computed_entry):
    """
    Generate a new ComputedStructureEntry for a molecule in a box, for the input elemental
    ComputedEntry
    """
    assert len(computed_entry.composition.elements) == 1  # Elemental!
    struc, formula, total_magnetization = make_molecule_in_a_box(
        computed_entry.data["pretty_formula"]
    )
    molecular_entry = ComputedStructureEntry(
        structure=struc,
        energy=computed_entry.energy_per_atom * 2,  # set entry energy to be hull energy
        composition=Composition(formula),
        parameters=None,
    )
    molecular_entry.data["oxide_type"] = "None"
    molecular_entry.data["pretty_formula"] = formula
    molecular_entry.data["e_above_hull"] = 0
    molecular_entry.data["band_gap"] = None
    molecular_entry.data["nsites"] = 2
    molecular_entry.data["volume"] = 0
    molecular_entry.data["icsd_id"] = None
    molecular_entry.data["formation_energy_per_atom"] = 0
    molecular_entry.data["energy_per_atom"] = computed_entry.data["energy_per_atom"]
    molecular_entry.data["energy"] = computed_entry.data["energy_per_atom"] * 2
    molecular_entry.data["total_magnetization"] = total_magnetization
    molecular_entry.data["nelements"] = 1
    molecular_entry.data["elements"] = [formula]
    molecular_entry.data["molecule"] = True

    return molecular_entry


def _calculate_formation_energies(data, elemental):
    df = pd.DataFrame(data)
    for d in data:
        for e in elemental.keys():
            d[e] = Composition(d["formula"]).as_dict()[e]

    df2 = pd.DataFrame(data)
    df2["formation_energy"] = df2["energy_per_fu"]
    for k, v in elemental.items():
        df2["formation_energy"] -= df2[k] * v

    df["formation_energy"] = df2["formation_energy"]

    df["num_atoms_in_fu"] = df["energy_per_fu"] / df["energy_per_atom"]
    df["num_species"] = df["formula"].apply(lambda x: len(Composition(x).as_dict()))

    # sort by num_species, then alphabetically, then by num_atoms_in_fu, then by formation_energy
    df.sort_values(
        by=["num_species", "formula", "num_atoms_in_fu", "formation_energy"],
        inplace=True,
    )
    # drop num_atoms_in_fu and num_species
    df.drop(columns=["num_atoms_in_fu", "num_species"], inplace=True)

    return df


def _renormalise_entry(entry, renormalisation_energy_per_atom):
    """
    Regenerate the input entry with an energy per atom decreased by renormalisation_energy_per_atom
    """
    renormalised_entry_dict = entry.as_dict().copy()
    renormalised_entry_dict[
        "energy"
    ] = entry.energy - renormalisation_energy_per_atom * sum(
        entry.composition.values()
    )  # entry.energy includes MP corrections as desired
    renormalised_entry = PDEntry.from_dict(renormalised_entry_dict)
    return renormalised_entry


def get_chempots_from_phase_diagram(bulk_ce, phase_diagram):
    """
    Get the chemical potential limits for the bulk computed entry in the supplied phase diagram.

    Args:
        bulk_ce: Pymatgen ComputedStructureEntry object for bulk entry / supercell
        phase_diagram: Pymatgen PhaseDiagram object for the system of interest
    """
    bulk_composition = bulk_ce.composition
    redcomp = bulk_composition.reduced_composition
    # append bulk_ce to phase diagram, if not present
    entries = phase_diagram.all_entries
    if not any(
        (ent.composition == bulk_ce.composition and ent.energy == bulk_ce.energy)
        for ent in entries
    ):
        entries.append(
            PDEntry(
                bulk_ce.composition,
                bulk_ce.energy,
                attribute="Bulk Material",
            )
        )
        phase_diagram = PhaseDiagram(entries)

    chem_lims = phase_diagram.get_all_chempots(redcomp)

    return chem_lims


class CompetingPhases:
    """
    Class to generate the input files for competing phases on the phase diagram for the host
    material (determining the chemical potential limits). Materials Project (MP) data is used,
    along with an uncertainty range specified by `e_above_hull`, to determine the relevant
    competing phases. Diatomic gaseous molecules are generated as molecules-in-a-box as appropriate.

    TODO: Need to add functionality to deal with cases where the bulk composition is not listed
    on the MP – warn user (i.e. check your sh*t) and generate the competing phases according to
    composition position within phase diagram.
    E.g. from pycdt chemical_potentials:
    #                 "However, no stable entry with this composition exists "
    #                 "in the MP database!\nPlease consider submitting the "
    #                 "POSCAR to the MP xtaltoolkit, so future users will "
    #                 "know about this structure:"
    #                 " https://materialsproject.org/#apps/xtaltoolkit\n" – see
    analyze_GGA_chempots code for example.
    TODO: Add note to notebook that if your bulk phase is lower energy than its version on the MP
    (e.g. distorted perovskite), then you should use this for your bulk competing phase calculation.
    """

    def __init__(
        self, composition, e_above_hull=0.1, api_key=None, full_phase_diagram=False
    ):
        """
        Args:
            composition (str, Composition): Composition of host material
                (e.g. 'LiFePO4', or Composition('LiFePO4'), or Composition({"Li":1, "Fe":1,
                "P":1, "O":4}))
            e_above_hull (float): Maximum energy-above-hull of Materials Project entries to be
                considered as competing phases. This is an uncertainty range for the
                MP-calculated formation energies, which may not be accurate due to functional
                choice (GGA vs hybrid DFT / GGA+U / RPA etc.), lack of vdW corrections etc.
                Any phases that (would) border the host material on the phase diagram, if their
                relative energy was downshifted by `e_above_hull`, are included.
                Default is 0.1 eV/atom.
            api_key (str): Materials Project (MP) API key, needed to access the MP database for
                competing phase generation. If not supplied, will attempt to read from
                environment variable `PMG_MAPI_KEY` (in `~/.pmgrc.yaml`) – see the `doped`
                homepage (https://github.com/SMTG-UCL/doped) for instructions on setting this up.
                This should correspond to the legacy MP API; from
                https://legacy.materialsproject.org/open.
            full_phase_diagram (bool): If True, include all phases on the MP phase diagram (
                with energy above hull < e_above_hull) for the chemical system of the input
                composition (not recommended). If False, only includes phases that (would) border
                the host material on the phase diagram (and thus set the chemical potential
                limits), if their relative energy was downshifted by `e_above_hull`.
                (Default is False).
        """
        self.api_key = api_key

        # create list of entries
        self._molecules_in_a_box = ["H2", "O2", "N2", "F2", "Cl2"]

        # TODO: Should hard code S (solid + S8), P, Te and Se in here too. Common anions with a
        #  lot of unnecessary polymorphs on MP
        # P_EaH=0 is red phosphorus (HSE06 groundstate), P_EaH=0.037 is black phosphorus (thermo
        # stable at RT), so only need to generate these

        # all data collected from materials project
        self.data = [
            "pretty_formula",
            "e_above_hull",
            "band_gap",
            "nsites",
            "volume",
            "icsd_id",
            "formation_energy_per_atom",
            "energy_per_atom",
            "energy",
            "total_magnetization",
            "nelements",
            "elements",
        ]

        # set bulk composition (Composition(Composition("LiFePO4")) = Composition("LiFePO4")))
        self.bulk_comp = Composition(composition)

        # test api_key:
        if api_key is not None:
            if len(api_key) == 32:
                raise ValueError(
                    "You are trying to use the new Materials Project (MP) API which is not "
                    "supported by doped. Please use the legacy MP API ("
                    "https://legacy.materialsproject.org/open)."
                )
            if 15 <= len(api_key) <= 20:
                self.eah = "e_above_hull"
            else:
                raise ValueError(
                    f"API key {api_key} is not a valid legacy Materials Project API key. These "
                    f"are available at https://legacy.materialsproject.org/open"
                )

        with MPRester(api_key=self.api_key) as mpr:
            self.MP_full_pd_entries = mpr.get_entries_in_chemsys(
                list(self.bulk_comp.as_dict().keys()),
                inc_structure="initial",
                property_data=self.data,
            )
        self.MP_full_pd_entries = [
            e for e in self.MP_full_pd_entries if e.data["e_above_hull"] <= e_above_hull
        ]
        self.MP_full_pd_entries.sort(
            key=lambda x: x.data["e_above_hull"]
        )  # sort by e_above_hull

        pd_entries = []
        # check that none of the elemental ones are molecules in a box
        for entry in self.MP_full_pd_entries.copy():
            if (
                entry.data["pretty_formula"] in self._molecules_in_a_box
                and entry.data["e_above_hull"] == 0
            ):
                # only first matching molecular entry
                molecular_entry = _make_molecular_entry(entry)
                if not any(
                    ent.data["molecule"]
                    and ent.data["pretty_formula"]
                    == molecular_entry.data["pretty_formula"]
                    for ent in pd_entries
                ):  # first entry only
                    pd_entries.append(molecular_entry)
                    self.MP_full_pd_entries.append(molecular_entry)

            elif entry.data["pretty_formula"] not in self._molecules_in_a_box:
                entry.data["molecule"] = False
                pd_entries.append(entry)

        pd_entries.sort(key=lambda x: x.energy_per_atom)  # sort by energy per atom
        phase_diagram = PhaseDiagram(pd_entries)
        bulk_entries = [
            entry
            for entry in pd_entries
            if entry.composition.reduced_composition
            == self.bulk_comp.reduced_composition
        ]
        bulk_ce = bulk_entries[
            0
        ]  # lowest energy entry for bulk composition (after sorting)
        self.MP_bulk_ce = bulk_ce

        if not full_phase_diagram:  # default
            # cull to only include any phases that would border the host material on the phase
            # diagram, if their relative energy was downshifted by `e_above_hull`:
            MP_gga_chempots = get_chempots_from_phase_diagram(bulk_ce, phase_diagram)

            MP_bordering_phases = {
                phase for facet in MP_gga_chempots for phase in facet.split("-")
            }
            self.entries = [
                entry
                for entry in pd_entries
                if entry.name in MP_bordering_phases or entry.is_element
            ]

            # add any phases that would border the host material on the phase diagram, if their
            # relative energy was downshifted by `e_above_hull`:
            for entry in pd_entries:
                if entry.name not in MP_bordering_phases and not entry.is_element:
                    # decrease entry energy per atom by `e_above_hull` eV/atom
                    renormalised_entry = _renormalise_entry(entry, e_above_hull)
                    new_phase_diagram = PhaseDiagram(
                        phase_diagram.entries + [renormalised_entry]
                    )
                    new_MP_gga_chempots = get_chempots_from_phase_diagram(
                        bulk_ce, new_phase_diagram
                    )

                    if new_MP_gga_chempots != MP_gga_chempots:
                        # new bordering phase, add to list
                        self.entries.append(entry)

        else:  # full_phase_diagram = True
            self.entries = pd_entries

        # sort entries by e_above_hull, and then by num_species, then alphabetically:
        self.entries.sort(
            key=lambda x: (
                x.data["e_above_hull"],
                len(Composition(x.name).as_dict()),
                x.name,
            )
        )

    # TODO: Similar refactor to work mainly off config dict object here as well (as vasp_input)?
    # TODO: Currently doesn't exactly match vaspup2.0 naming convention which means it doesn't
    #  account for the ordering switch from 1..9 to 10 etc
    def convergence_setup(
        self,
        kpoints_metals=(40, 120, 5),
        kpoints_nonmetals=(5, 80, 5),
        user_potcar_functional="PBE_54",
        user_potcar_settings=None,
        user_incar_settings=None,
        **kwargs,
    ):
        """
        Generates VASP input files for kpoints convergence testing for competing phases,
        using PBEsol (GGA) DFT by default. Automatically sets the `ISMEAR` `INCAR` tag to 2 (if
        metallic) or 0 if not. Recommended to use with https://github.com/kavanase/vaspup2.0.

        Args:
            kpoints_metals (tuple): Kpoint density per inverse volume (Å^-3) to be tested in
                (min, max, step) format for metals
            kpoints_nonmetals (tuple): Kpoint density per inverse volume (Å^-3) to be tested in
                (min, max, step) format for nonmetals
            user_potcar_functional (str): POTCAR functional to use (default = "PBE_54")
            user_potcar_settings (dict): Override the default POTCARs, e.g. {"Li": "Li_sv"}. See
                `doped/PotcarSet.yaml` for the default `POTCAR` set.
            user_incar_settings (dict): Override the default INCAR settings
                e.g. {"EDIFF": 1e-5, "LDAU": False, "ALGO": "All"}. Note that any non-numerical or
                non-True/False flags need to be input as strings with quotation marks. See
                `doped/PBEsol_ConvergenceSet.yaml` for the default settings.
        """
        # by default uses pbesol, but easy to switch to pbe or pbe+u using user_incar_settings
        pbesol_convrg_set = loadfn(
            os.path.join(MODULE_DIR, "PBEsol_ConvergenceSet.yaml")
        )

        # kpoints should be set as (min, max, step)
        min_nm, max_nm, step_nm = kpoints_nonmetals
        min_m, max_m, step_m = kpoints_metals

        potcar_dict = copy.deepcopy(default_potcar_dict)
        if user_potcar_settings:
            potcar_dict["POTCAR"].update(user_potcar_settings)
        if user_potcar_functional:
            potcar_dict["POTCAR_FUNCTIONAL"] = user_potcar_functional
        pbesol_convrg_set.update(potcar_dict)

        # separate metals and non-metals
        self.nonmetals = []
        self.metals = []
        for e in self.entries:
            if not e.data["molecule"]:
                if e.data["band_gap"] > 0:
                    self.nonmetals.append(e)
                else:
                    self.metals.append(e)
            else:
                print(
                    f"{e.name} is a molecule in a box, does not need convergence testing"
                )

        for e in self.nonmetals:
            if user_incar_settings is not None:
                uis = copy.deepcopy(user_incar_settings)
            else:
                uis = {}
            if e.data["total_magnetization"] > 0.1:  # account for magnetic moment
                if "ISPIN" not in uis:
                    uis["ISPIN"] = 2
                if "NUPDOWN" not in uis and int(e.data["total_magnetization"]) > 0:
                    uis["NUPDOWN"] = int(e.data["total_magnetization"])

            for kpoint in range(min_nm, max_nm, step_nm):
                dis = DictSet(
                    e.structure,
                    pbesol_convrg_set,
                    user_kpoints_settings={"reciprocal_density": kpoint},
                    user_incar_settings=uis,
                    force_gamma=True,
                )

                kname = "k" + ",".join(str(k) for k in dis.kpoints.kpts[0])
                fname = (
                    f"competing_phases/{e.name}_EaH"
                    f"_{round(e.data['e_above_hull'],4)}/kpoint_converge/{kname}"
                )  # TODO: competing_phases folder name should be an optional parameter
                dis.write_input(fname, **kwargs)

        for e in self.metals:
            if user_incar_settings is not None:
                uis = copy.deepcopy(user_incar_settings)
            else:
                uis = {}
            # change the ismear and sigma for metals
            uis["ISMEAR"] = -5
            uis["SIGMA"] = 0.2

            if e.data["total_magnetization"] > 0.1:  # account for magnetic moment
                if "ISPIN" not in uis:
                    uis["ISPIN"] = 2
                if "NUPDOWN" not in uis and int(e.data["total_magnetization"]) > 0:
                    uis["NUPDOWN"] = int(e.data["total_magnetization"])

            for kpoint in range(min_m, max_m, step_m):
                dis = DictSet(
                    e.structure,
                    pbesol_convrg_set,
                    user_kpoints_settings={"reciprocal_density": kpoint},
                    user_incar_settings=uis,
                    force_gamma=True,
                )

                kname = "k" + ",".join(str(k) for k in dis.kpoints.kpts[0])
                fname = (
                    f"competing_phases/{e.name}_EaH_"
                    f"{round(e.data['e_above_hull'],4)}/kpoint_converge/{kname}"
                )
                dis.write_input(fname, **kwargs)

    # TODO: Add vasp_ncl_setup()
    def vasp_std_setup(
        self,
        kpoints_metals=95,
        kpoints_nonmetals=45,
        user_potcar_functional="PBE_54",
        user_potcar_settings=None,
        user_incar_settings=None,
        **kwargs,
    ):
        """
        Generates VASP input files for `vasp_std` relaxations of the competing phases,
        using HSE06 (hybrid DFT) DFT by default. Automatically sets the `ISMEAR` `INCAR` tag to 2
        (if metallic) or 0 if not. Note that any changes to the default `INCAR`/`POTCAR` settings
        should be consistent with those used for the defect supercell calculations.

        Args:
            kpoints_metals (int): Kpoint density per inverse volume (Å^-3) for metals
            kpoints_nonmetals (int): Kpoint density per inverse volume (Å^-3) for nonmetals
            user_potcar_functional (str): POTCAR functional to use (default = "PBE_54")
            user_potcar_settings (dict): Override the default POTCARs, e.g. {"Li": "Li_sv"}. See
                `doped/PotcarSet.yaml` for the default `POTCAR` set.
            user_incar_settings (dict): Override the default INCAR settings
                e.g. {"EDIFF": 1e-5, "LDAU": False, "ALGO": "All"}. Note that any non-numerical or
                non-True/False flags need to be input as strings with quotation marks. See
                `doped/HSE06_RelaxSet.yaml` for the default settings.
        """
        # TODO: Update this to use:
        #  sym = SpacegroupAnalyzer(e.structure)
        #  struc = sym.get_primitive_standard_structure() -> output this structure
        hse06_relax_set = loadfn(os.path.join(MODULE_DIR, "HSE06_RelaxSet.yaml"))

        potcar_dict = copy.deepcopy(default_potcar_dict)
        if user_potcar_settings:
            potcar_dict["POTCAR"].update(user_potcar_settings)
        if user_potcar_functional:
            potcar_dict["POTCAR_FUNCTIONAL"] = user_potcar_functional
        hse06_relax_set.update(potcar_dict)

        # separate metals, non-metals and molecules
        self.nonmetals = []
        self.metals = []
        self.molecules = []
        for e in self.entries:
            if e.data["molecule"]:
                self.molecules.append(e)
            else:
                if e.data["band_gap"] > 0:
                    self.nonmetals.append(e)
                else:
                    self.metals.append(e)

        for e in self.nonmetals:
            if user_incar_settings is not None:
                uis = copy.deepcopy(user_incar_settings)
            else:
                uis = {}
            if e.data["total_magnetization"] > 0.1:  # account for magnetic moment
                if "ISPIN" not in uis:
                    uis["ISPIN"] = 2
                if "NUPDOWN" not in uis and int(e.data["total_magnetization"]) > 0:
                    uis["NUPDOWN"] = int(e.data["total_magnetization"])

            dis = DictSet(
                e.structure,
                hse06_relax_set,
                user_kpoints_settings={"reciprocal_density": kpoints_nonmetals},
                user_incar_settings=uis,
                force_gamma=True,
            )

            fname = (
                f"competing_phases/{e.name}_EaH_"
                f"{round(e.data['e_above_hull'],4)}/vasp_std"
            )
            dis.write_input(fname, **kwargs)

        for e in self.metals:
            if user_incar_settings is not None:
                uis = copy.deepcopy(user_incar_settings)
            else:
                uis = {}
            # change the ismear and sigma for metals
            uis["ISMEAR"] = 1
            uis["SIGMA"] = 0.2

            if e.data["total_magnetization"] > 0.1:  # account for magnetic moment
                if "ISPIN" not in uis:
                    uis["ISPIN"] = 2
                if "NUPDOWN" not in uis and int(e.data["total_magnetization"]) > 0:
                    uis["NUPDOWN"] = int(e.data["total_magnetization"])

            dis = DictSet(
                e.structure,
                hse06_relax_set,
                user_kpoints_settings={"reciprocal_density": kpoints_metals},
                user_incar_settings=uis,
                force_gamma=True,
            )
            fname = (
                f"competing_phases/{e.name}_EaH_"
                f"{round(e.data['e_above_hull'],4)}/vasp_std"
            )
            dis.write_input(fname, **kwargs)

        for e in self.molecules:  # gamma-only for molecules
            if user_incar_settings is not None:
                uis = copy.deepcopy(user_incar_settings)
            else:
                uis = {}

            uis["ISIF"] = 2  # can't change the volume

            if e.data["total_magnetization"] > 0.1:  # account for magnetic moment
                if "ISPIN" not in uis:
                    uis["ISPIN"] = 2
                if "NUPDOWN" not in uis and int(e.data["total_magnetization"]) > 0:
                    uis["NUPDOWN"] = int(e.data["total_magnetization"])

            dis = DictSet(
                e.structure,
                hse06_relax_set,
                user_kpoints_settings=Kpoints().from_dict(
                    {
                        "comment": "Gamma-only kpoints for molecule-in-a-box",
                        "generation_style": "Gamma",
                    }
                ),
                user_incar_settings=uis,
                force_gamma=True,
            )
            fname = f"competing_phases/{e.name}_EaH_{round(e.data['e_above_hull'],4)}/vasp_std"
            dis.write_input(fname, **kwargs)


# TODO: Add full_sub_approach option
# TODO: Add warnings for full_sub_approach=True, especially if done with multiple
#  extrinsic species.
class ExtrinsicCompetingPhases(CompetingPhases):
    """
    This class generates the competing phases that need to be calculated to obtain the chemical
    potential limits when doping with extrinsic species / impurities. Ensures that only the
    necessary additional competing phases are generated.
    """

    def __init__(
        self,
        composition,
        extrinsic_species,
        e_above_hull=0.1,
        full_sub_approach=False,
        codoping=False,
        api_key=None,
    ):
        """
        Args:
            composition (str, Composition): Composition of host material
                (e.g. 'LiFePO4', or Composition('LiFePO4'), or Composition({"Li":1, "Fe":1,
                "P":1, "O":4}))
            extrinsic_species (str, list): Extrinsic dopant/impurity species
                (e.g. "Mg" or ["Mg", "Na"])
            e_above_hull (float): Maximum energy-above-hull of Materials Project entries to be
                considered as competing phases. This is an uncertainty range for the
                MP-calculated formation energies, which may not be accurate due to functional
                choice (GGA vs hybrid DFT / GGA+U / RPA etc.), lack of vdW corrections etc.
                Any phases that would border the host material on the phase diagram, if their
                relative energy was downshifted by `e_above_hull`, are included.
                Default is 0.1 eV/atom.
            full_sub_approach (bool): Generate competing phases by considering the full phase
                diagram, including chemical potential limits with multiple extrinsic phases.
                Only recommended when looking at high (non-dilute) doping concentrations.
                Default = False. Described in further detail below.
            codoping (bool): Whether to consider extrinsic competing phases containing multiple
                extrinsic species. Only relevant to high (non-dilute) co-doping concentrations.
                If set to True, then `full_sub_approach` is also set to True.
                Default = False.
            api_key (str): Materials Project (MP) API key, needed to access the MP database for
                competing phase generation. If not supplied, will attempt to read from
                environment variable `PMG_MAPI_KEY` (in `~/.pmgrc.yaml`) – see the `doped`
                homepage (https://github.com/SMTG-UCL/doped) for instructions on setting this up.
                This should correspond to the legacy MP API; from
                https://legacy.materialsproject.org/open.

        This code uses the Materials Project (MP) phase diagram data along with the
        `e_above_hull` error range to generate potential competing phases.

        NOTE on 'full_sub_approach':
            The default approach for substitutional elements (`full_sub_approach = False`) is to
            only consider chemical potential limits with a maximum of 1 extrinsic phase
            (composition with extrinsic species present). This is a valid approximation for the
            case of dilute dopant/impurity concentrations. For high (non-dilute) concentrations
            of extrinsic species, use `full_sub_approach = True`.
        """
        # the competing phases & entries of the OG system
        super().__init__(composition, e_above_hull, api_key)
        self.intrinsic_entries = copy.deepcopy(self.entries)
        self.entries = []
        self.intrinsic_species = [
            s.symbol for s in self.bulk_comp.reduced_composition.elements
        ]
        self.MP_intrinsic_full_pd_entries = (
            self.MP_full_pd_entries
        )  # includes molecules-in-boxes

        if isinstance(extrinsic_species, str):
            extrinsic_species = [
                extrinsic_species,
            ]
        elif not isinstance(extrinsic_species, list):
            raise TypeError(
                f"`extrinsic_species` must be a string (i.e. the extrinsic species "
                f"symbol, e.g. 'Mg') or a list (e.g. ['Mg', 'Na']), got type "
                f"{type(extrinsic_species)} instead!"
            )
        self.extrinsic_species = extrinsic_species

        # if codoping = True, should have multiple extrinsic species
        if codoping:
            if len(extrinsic_species) < 2:
                warnings.warn(
                    "`codoping` is set to True, but `extrinsic_species` only contains 1 "
                    "element, so `codoping` will be set to False."
                )
                codoping = False

            elif not full_sub_approach:
                full_sub_approach = True

        if (
            full_sub_approach
        ):  # can be time-consuming if several extrinsic_species supplied
            if codoping:
                # TODO: `full_sub_approach` shouldn't necessarily mean `full_phase_diagram =
                #  True` right? As in can be non-full-phase-diagram intrinsic + extrinsic
                #  entries, including facets with multiple extrinsic entries but still not the
                #  full phase diagram? – To be updated!
                # TODO: When `full_phase_diagram` option added to `CompetingPhases`, can remove
                #  this code block and just use:
                #  super()__init__(composition = (
                #  self.intrinsic_species+self.extrinsic_species).join(""), e_above_hull, api_key)
                #  )
                #  self.intrinsic_entries = [phase for phase in self.entries if
                #  not any([extrinsic in phase for extrinsic in self.extrinsic_species])]
                #  entries = [phase for phase in self.entries if phase not in
                #  self.intrinsic_entries]
                #  self.entries = entries
                #  self.MP_intrinsic_full_pd_entries = [entry for entry in
                #  self.MP_full_pd_entries if not any([extrinsic in entry.composition.reduced_formula
                #  for extrinsic in self.extrinsic_species])]
                #  MP_full_pd_entries = [entry for entry in self.MP_full_pd_entries if entry not in
                #  self.MP_intrinsic_full_pd_entries]
                #  self.MP_full_pd_entries = MP_full_pd_entries  # includes molecules-in-boxes

                with MPRester(api_key=self.api_key) as mpr:
                    self.MP_full_pd_entries = mpr.get_entries_in_chemsys(
                        self.intrinsic_species + self.extrinsic_species,
                        inc_structure="initial",
                        property_data=self.data,
                    )
                self.MP_full_pd_entries = [
                    e
                    for e in self.MP_full_pd_entries
                    if e.data["e_above_hull"] <= e_above_hull
                ]

                # sort by e_above_hull:
                self.MP_full_pd_entries.sort(key=lambda x: x.data["e_above_hull"])

                for entry in self.MP_full_pd_entries.copy():
                    if any(
                        sub_elt in entry.composition
                        for sub_elt in self.extrinsic_species
                    ):
                        if (
                            entry.data["pretty_formula"] in self._molecules_in_a_box
                            and entry.data["e_above_hull"] == 0
                        ):  # only first matching entry
                            molecular_entry = _make_molecular_entry(entry)
                            if not any(
                                entry.data["molecule"]
                                and entry.data["pretty_formula"]
                                == molecular_entry.data["pretty_formula"]
                                for entry in self.entries
                            ):  # first entry only
                                self.MP_full_pd_entries.append(molecular_entry)
                                self.entries.append(molecular_entry)
                        elif (
                            entry.data["pretty_formula"] not in self._molecules_in_a_box
                        ):
                            entry.data["molecule"] = False
                            self.entries.append(entry)

                # sort entries by e_above_hull, and then by num_species, then alphabetically:
                self.entries.sort(
                    key=lambda x: (
                        x.data["e_above_hull"],
                        len(Composition(x.name).as_dict()),
                        x.name,
                    )
                )

            else:  # full_sub_approach but not co-doping
                self.MP_full_pd_entries = []
                for sub_elt in self.extrinsic_species:
                    extrinsic_entries = []
                    with MPRester(api_key=self.api_key) as mpr:
                        MP_full_pd_entries = mpr.get_entries_in_chemsys(
                            self.intrinsic_species
                            + [
                                sub_elt,
                            ],
                            inc_structure="initial",
                            property_data=self.data,
                        )
                    MP_full_pd_entries = [
                        e
                        for e in MP_full_pd_entries
                        if e.data["e_above_hull"] <= e_above_hull
                    ]
                    # sort by e_above_hull:
                    MP_full_pd_entries.sort(key=lambda x: x.data["e_above_hull"])

                    for entry in MP_full_pd_entries.copy():
                        if entry not in self.MP_full_pd_entries:
                            self.MP_full_pd_entries.append(entry)
                        if sub_elt in entry.composition:
                            if (
                                entry.data["pretty_formula"] in self._molecules_in_a_box
                                and entry.data["e_above_hull"] == 0
                            ):
                                # only first matching entry
                                molecular_entry = _make_molecular_entry(entry)
                                if not any(
                                    entry.data["molecule"]
                                    and entry.data["pretty_formula"]
                                    == molecular_entry.data["pretty_formula"]
                                    for entry in self.entries
                                ):  # first entry only
                                    self.MP_full_pd_entries.append(molecular_entry)
                                    extrinsic_entries.append(molecular_entry)

                            elif (
                                entry.data["pretty_formula"]
                                not in self._molecules_in_a_box
                            ):
                                entry.data["molecule"] = False
                                extrinsic_entries.append(entry)

                            # sort entries by e_above_hull, and then by num_species,
                            # then alphabetically:
                            extrinsic_entries.sort(
                                key=lambda x: (
                                    x.data["e_above_hull"],
                                    len(Composition(x.name).as_dict()),
                                    x.name,
                                )
                            )
                            self.entries += extrinsic_entries

        else:  # full_sub_approach = False; recommended approach for extrinsic species (assumes
            # dilute concentrations)

            # now compile substitution entries:
            self.MP_full_pd_entries = []

            for sub_elt in self.extrinsic_species:
                extrinsic_pd_entries = []
                with MPRester(api_key=self.api_key) as mpr:
                    MP_full_pd_entries = mpr.get_entries_in_chemsys(
                        self.intrinsic_species
                        + [
                            sub_elt,
                        ],
                        inc_structure="initial",
                        property_data=self.data,
                    )
                self.MP_full_pd_entries = [
                    e
                    for e in MP_full_pd_entries
                    if e.data["e_above_hull"] <= e_above_hull
                ]
                # sort by e_above_hull:
                self.MP_full_pd_entries.sort(key=lambda x: x.data["e_above_hull"])

                for entry in self.MP_full_pd_entries.copy():
                    if (
                        entry.data["pretty_formula"] in self._molecules_in_a_box
                        and entry.data["e_above_hull"] == 0
                    ):
                        # only first matching entry
                        molecular_entry = _make_molecular_entry(entry)
                        if not any(
                            entry.data["molecule"]
                            and entry.data["pretty_formula"]
                            == molecular_entry.data["pretty_formula"]
                            for entry in extrinsic_pd_entries
                        ):  # first entry only
                            self.MP_full_pd_entries.append(molecular_entry)
                            extrinsic_pd_entries.append(molecular_entry)
                    elif entry.data["pretty_formula"] not in self._molecules_in_a_box:
                        entry.data["molecule"] = False
                        extrinsic_pd_entries.append(entry)

                # Adding substitutional phases to extrinsic competing phases list only when the
                # phases in equilibria are those from the bulk phase diagram. This is essentially
                # the assumption that the majority of elements in the total composition will be
                # from the host composition rather than the extrinsic species (a good
                # approximation for dilute concentrations)

                if not extrinsic_pd_entries:
                    raise ValueError(
                        f"No Materials Project entries found for the given chemical "
                        f"system: {self.intrinsic_species + [sub_elt,]}"
                    )

                extrinsic_phase_diagram = PhaseDiagram(extrinsic_pd_entries)
                MP_extrinsic_gga_chempots = get_chempots_from_phase_diagram(
                    self.MP_bulk_ce, extrinsic_phase_diagram
                )
                MP_extrinsic_bordering_phases = []

                for facet in MP_extrinsic_gga_chempots.keys():
                    # if the number of intrinsic competing phases for this facet is equal to the
                    # number of species in the bulk composition, then include the extrinsic phase(s)
                    # for this facet (full_sub_approach = False approach)
                    MP_intrinsic_bordering_phases = {
                        phase for phase in facet.split("-") if sub_elt not in phase
                    }
                    if len(MP_intrinsic_bordering_phases) == len(
                        self.intrinsic_species
                    ):
                        # add to list of extrinsic bordering phases, if not already present:
                        MP_extrinsic_bordering_phases.extend(
                            [
                                phase
                                for phase in facet.split("-")
                                if sub_elt in phase
                                and phase not in MP_extrinsic_bordering_phases
                            ]
                        )

                # add any phases that would border the host material on the phase diagram,
                # if their relative energy was downshifted by `e_above_hull`:
                for entry in MP_full_pd_entries:
                    if (
                        entry.name not in MP_extrinsic_bordering_phases
                        and not entry.is_element
                        and sub_elt in entry.composition
                    ):
                        # decrease entry energy per atom by `e_above_hull` eV/atom
                        renormalised_entry = _renormalise_entry(entry, e_above_hull)
                        new_extrinsic_phase_diagram = PhaseDiagram(
                            extrinsic_phase_diagram.entries + [renormalised_entry]
                        )
                        new_MP_extrinsic_gga_chempots = get_chempots_from_phase_diagram(
                            self.MP_bulk_ce, new_extrinsic_phase_diagram
                        )

                        if new_MP_extrinsic_gga_chempots != MP_extrinsic_gga_chempots:
                            # new bordering phase, check if not an over-dependent facet:

                            for facet in new_MP_extrinsic_gga_chempots:
                                if facet not in MP_extrinsic_gga_chempots:
                                    # new facet, check if not an over-dependent facet:
                                    MP_intrinsic_bordering_phases = {
                                        phase
                                        for phase in facet.split("-")
                                        if sub_elt not in phase
                                    }
                                    if len(MP_intrinsic_bordering_phases) == len(
                                        self.intrinsic_species
                                    ):
                                        MP_extrinsic_bordering_phases.extend(
                                            [
                                                phase
                                                for phase in facet.split("-")
                                                if sub_elt in phase
                                                and phase
                                                not in MP_extrinsic_bordering_phases
                                            ]
                                        )

                extrinsic_entries = [
                    entry
                    for entry in extrinsic_pd_entries
                    if entry.name in MP_extrinsic_bordering_phases
                    or (entry.is_element and sub_elt in entry.name)
                ]

                # check that extrinsic competing phases list is not empty (can happen with
                # 'over-dependent' facets); if so then set full_sub_approach = True and re-run
                # the extrinsic phase addition process
                if not extrinsic_entries:
                    warnings.warn(
                        "Determined chemical potentials to be over dependent on an "
                        "extrinsic species. This means we need to revert to "
                        "`full_sub_approach = True` – running now."
                    )
                    full_sub_approach = True
                    extrinsic_entries = [
                        entry
                        for entry in self.MP_full_pd_entries
                        if sub_elt in entry.composition
                    ]

                # sort entries by e_above_hull, and then by num_species, then alphabetically:
                extrinsic_entries.sort(
                    key=lambda x: (
                        x.data["e_above_hull"],
                        len(Composition(x.name).as_dict()),
                        x.name,
                    )
                )
                self.entries += extrinsic_entries


class CompetingPhasesAnalyzer:
    """
    Post processing competing phases data to calculate chemical potentials.
    """

    def __init__(self, system, extrinsic_species=None):
        """
        Args:
            system (str): The  'reduced formula' of the bulk composition
            extrinsic_species (str): Dopant species - can only deal with one at a time (see
            notebook in examples folder for more complex cases)
        """

        self.bulk_composition = Composition(system)
        self.elemental = [str(c) for c in self.bulk_composition.elements]
        self.extrinsic_species = extrinsic_species

        if extrinsic_species:
            self.elemental.append(extrinsic_species)

    def from_vaspruns(self, path="competing_phases", folder="vasp_std", csv_fname=None):
        """
        Reads in vaspruns, collates energies to csv.

        Args:
            path (list, str, pathlib Path): Either a list of strings or Paths to vasprun.xml(.gz)
            files, or a path to the base folder in which you have your
            formula_EaH_/vasp_std/vasprun.xml
            folder (str): The folder in which vasprun is, only use if you set base path
            (i.e. change to vasp_ncl or if vaspruns are in the formula_EaH folder).
            csv_fname (str): If set will save to csv with this name
        Returns:
            None, sets self.data and self.elemental_energies
        """
        # TODO: Change this to just recursively search for vaspruns within the specified path
        # TODO: Add check for matching INCAR and POTCARs from these calcs, as we also want with
        #  defect parsing
        self.vasprun_paths = []
        # fetch data
        # if path is just a list of all competing phases
        if isinstance(path, list):
            for p in path:
                if "vasprun.xml" in Path(p).name:
                    self.vasprun_paths.append(str(Path(p)))

                # try to find the file - will always pick the first match for vasprun.xml*
                elif len(list(Path(p).glob("vasprun.xml*"))) > 0:
                    vsp = list(Path(p).glob("vasprun.xml*"))[0]
                    self.vasprun_paths.append(str(vsp))

                else:
                    print(
                        f"Can't find a vasprun.xml(.gz) file for {p}, proceed with caution"
                    )

        # if path provided points to the doped created directories
        elif isinstance(path, (PurePath, str)):
            path = Path(path)
            for p in path.iterdir():
                if p.glob("EaH"):
                    vp = p / folder / "vasprun.xml"
                    try:
                        vr, vr_path = get_vasprun(vp)
                        self.vasprun_paths.append(vr_path)

                    except FileNotFoundError:
                        try:
                            vp = p / "vasprun.xml"
                            vr, vr_path = get_vasprun(vp)
                            self.vasprun_paths.append(vr_path)

                        except FileNotFoundError:
                            print(
                                f"Can't find a vasprun.xml(.gz) file in {p} or {p/folder}, "
                                f"proceed with caution"
                            )
                            continue

                else:
                    raise FileNotFoundError(
                        "Folders are not in the correct structure, provide them as a list of "
                        "paths (or strings)"
                    )

        else:
            raise ValueError(
                "Path should either be a list of paths, a string or a pathlib Path object"
            )

        # Ignore POTCAR warnings when loading vasprun.xml
        # pymatgen assumes the default PBE with no way of changing this
        warnings.filterwarnings("ignore", category=UnknownPotcarWarning)
        warnings.filterwarnings(
            "ignore", message="No POTCAR file with matching TITEL fields"
        )

        num = len(self.vasprun_paths)
        print(
            f"Parsing {num} vaspruns and pruning to include only lowest-energy polymorphs..."
        )
        self.vaspruns = [Vasprun(e).as_dict() for e in self.vasprun_paths]
        self.data = []

        temp_data = []
        self.elemental_energies = {}
        for v in self.vaspruns:
            rcf = v["reduced_cell_formula"]
            comp = Composition(v["unit_cell_formula"])
            formulas_per_unit = comp.get_reduced_composition_and_factor()[1]
            final_energy = v["output"]["final_energy"]
            kpoints = "x".join(str(x) for x in v["input"]["kpoints"]["kpoints"][0])

            # check if elemental:
            if len(rcf) == 1:
                elt = v["elements"][0]
                if elt not in self.elemental_energies:
                    self.elemental_energies[elt] = v["output"]["final_energy_per_atom"]
                    if elt not in self.elemental:  # new (extrinsic) element
                        self.extrinsic_species = elt
                        self.elemental.append(elt)

                elif (
                    v["output"]["final_energy_per_atom"] < self.elemental_energies[elt]
                ):
                    # only include lowest energy elemental polymorph
                    self.elemental_energies[elt] = v["output"]["final_energy_per_atom"]

            d = {
                "formula": v["pretty_formula"],
                "kpoints": kpoints,
                "energy_per_fu": final_energy / formulas_per_unit,
                "energy_per_atom": v["output"]["final_energy_per_atom"],
                "energy": final_energy,
            }
            temp_data.append(d)

        df = _calculate_formation_energies(temp_data, self.elemental_energies)
        if csv_fname is not None:
            df.to_csv(csv_fname, index=False)
            print(f"Competing phase formation energies have been saved to {csv_fname}.")

        self.data = df.to_dict(orient="records")

    def from_csv(self, csv):
        """
        Read in data from csv.
        Args:
            csv (str): Path to csv file. Must have columns 'formula',
                'energy_per_fu', 'energy' and 'formation_energy'
        Returns:
            None, sets self.data and self.elemental_energies
        """
        df = pd.read_csv(csv)
        columns = [
            "formula",
            "energy_per_fu",
            "energy_per_atom",
            "energy",
            "formation_energy",
        ]
        if all(x in list(df.columns) for x in columns):
            droplist = [i for i in df.columns if i not in columns]
            df.drop(droplist, axis=1, inplace=True)
            d = df.to_dict(orient="records")
            self.data = d

            self.elemental_energies = {}
            for i in d:
                c = Composition(i["formula"])
                if len(c.elements) == 1:
                    elt = c.chemical_system
                    if (
                        elt not in self.elemental_energies
                        or i["energy_per_atom"] < self.elemental_energies[elt]
                    ):
                        self.elemental_energies[elt] = i["energy_per_atom"]

        else:
            raise ValueError(
                "supplied csv does not contain the correct headers, cannot read in the data"
            )

    def calculate_chempots(self, csv_fname=None, verbose=True):
        """
        Calculates chemcial potential limits. For dopant species, it calculates the limiting
        potential based on the intrinsic chemical potentials (i.e. same as
        `full_sub_approach=False` in pycdt)
        Args:
            csv_fname (str): If set, will save chemical potential limits to csv
            verbose (bool): If True, will print out chemical potential limits.
        Retruns:
            Pandas DataFrame, optionally saved to csv
        """

        intrinsic_phase_diagram_entries = []
        extrinsic_formation_energies = []
        bulk_pde_list = []
        for d in self.data:
            e = PDEntry(d["formula"], d["energy_per_fu"])
            # presumably checks if the phase is intrinsic
            if set(Composition(d["formula"]).elements).issubset(
                self.bulk_composition.elements
            ):
                intrinsic_phase_diagram_entries.append(e)
                if e.composition == self.bulk_composition:  # bulk phase
                    bulk_pde_list.append(e)
            else:
                extrinsic_formation_energies.append(
                    {"formula": d["formula"], "formation_energy": d["formation_energy"]}
                )

        if len(bulk_pde_list) == 0:
            raise ValueError(
                f"Could not find bulk phase for "
                f"{self.bulk_composition.reduced_formula} in the supplied data. "
                f"Found phases: "
                f"{ {e.composition.reduced_formula for e in intrinsic_phase_diagram_entries} }"
            )
        if len(bulk_pde_list) > 0:
            # lowest energy bulk phase
            self.bulk_pde = sorted(bulk_pde_list, key=lambda x: x.energy_per_atom)[0]

        self._intrinsic_phase_diagram = PhaseDiagram(
            intrinsic_phase_diagram_entries,
            map(Element, self.bulk_composition.elements),
        )

        # check if it's stable and if not error out
        if self.bulk_pde not in self._intrinsic_phase_diagram.stable_entries:
            eah = self._intrinsic_phase_diagram.get_e_above_hull(self.bulk_pde)
            raise ValueError(
                f"{self.bulk_composition.reduced_formula} is not stable with respect to competing "
                f"phases, EaH={eah:.4f} eV/atom"
            )

        chem_lims = self._intrinsic_phase_diagram.get_all_chempots(
            self.bulk_composition
        )
        # remove Element to make it jsonable
        no_element_chem_lims = {}
        for k, v in chem_lims.items():
            temp_dict = {}
            for kk, vv in v.items():
                temp_dict[str(kk)] = vv
            no_element_chem_lims[k] = temp_dict

        self._intrinsic_chem_limits = {
            "facets": no_element_chem_lims,
            "elemental_refs": {
                str(elt): ent.energy_per_atom
                for elt, ent in self._intrinsic_phase_diagram.el_refs.items()
            },
            "facets_wrt_el_refs": {},
        }

        # relate the facets to the elemental energies
        for facet, chempot_dict in self._intrinsic_chem_limits["facets"].items():
            relative_chempot_dict = copy.deepcopy(chempot_dict)
            for e in relative_chempot_dict.keys():
                relative_chempot_dict[e] -= self._intrinsic_chem_limits[
                    "elemental_refs"
                ][e]
            self._intrinsic_chem_limits["facets_wrt_el_refs"].update(
                {facet: relative_chempot_dict}
            )

        # get chemical potentials as pandas dataframe
        chemical_potentials = []
        for k, v in self._intrinsic_chem_limits["facets_wrt_el_refs"].items():
            lst = []
            columns = []
            for k, v in v.items():
                lst.append(v)
                columns.append(str(k))
            chemical_potentials.append(lst)

        # make df, will need it in next step
        df = pd.DataFrame(chemical_potentials, columns=columns)

        if self.extrinsic_species is not None:
            if verbose:
                print(f"Calculating chempots for {self.extrinsic_species}")
            for e in extrinsic_formation_energies:
                for el in self.elemental:
                    e[el] = Composition(e["formula"]).as_dict()[el]

            # gets the df into a slightly more convenient dict
            cpd = df.to_dict(orient="records")
            mins = []
            mins_formulas = []
            df3 = pd.DataFrame(extrinsic_formation_energies)
            for i, c in enumerate(cpd):
                name = f"mu_{self.extrinsic_species}_{i}"
                df3[name] = df3["formation_energy"]
                for k, v in c.items():
                    df3[name] -= df3[k] * v
                df3[name] /= df3[self.extrinsic_species]
                # find min at that chempot
                mins.append(df3[name].min())
                mins_formulas.append(df3.iloc[df3[name].idxmin()]["formula"])

            df[self.extrinsic_species] = mins
            col_name = f"{self.extrinsic_species}_limiting_phase"
            df[col_name] = mins_formulas

            # 1. work out the formation energies of all dopant competing
            #    phases using the elemental energies
            # 2. for each of the chempots already calculated work out what
            #    the chemical potential of the dopant would be from
            #       mu_dopant = Hf(dopant competing phase) - sum(mu_elements)
            # 3. find the most negative mu_dopant which then becomes the new
            #    canonical chemical potential for that dopant species and the
            #    competing phase is the 'limiting phase' right?
            # 4. update the chemical potential limits table to reflect this?
            # TODO: I don't think this is right. Here it's just getting the extrinsic chempots at
            #  the intrinsic chempot limits, but actually it should be recalculating the chempot
            #  limits with the extrinsic competing phases, then reducing _these_ facets down to
            #  those with only one extrinsic competing phase bordering

            # reverse engineer chem lims for extrinsic
            df4 = df.copy().to_dict(orient="records")
            cl2 = {
                "elemental_refs": self.elemental_energies,
                "facets_wrt_el_refs": {},
                "facets": {},
            }

            for i, d in enumerate(df4):
                key = (
                    list(self._intrinsic_chem_limits["facets_wrt_el_refs"].keys())[i]
                    + "-"
                    + d[col_name]
                )
                new_vals = list(
                    self._intrinsic_chem_limits["facets_wrt_el_refs"].values()
                )[i]
                new_vals[f"{self.extrinsic_species}"] = d[f"{self.extrinsic_species}"]
                cl2["facets_wrt_el_refs"][key] = new_vals

            # relate the facets to the elemental
            # energies but in reverse this time
            for facet, chempot_dict in cl2["facets_wrt_el_refs"].items():
                relative_chempot_dict = copy.deepcopy(chempot_dict)
                for e in relative_chempot_dict.keys():
                    relative_chempot_dict[e] += cl2["elemental_refs"][e]
                cl2["facets"].update({facet: relative_chempot_dict})

            self._chem_limits = cl2

        else:  # intrinsic only
            self._chem_limits = self._intrinsic_chem_limits

        # save and print
        if csv_fname is not None:
            df.to_csv(csv_fname, index=False)
            if verbose:
                print("Saved chemical potential limits to csv file: ", csv_fname)

        if verbose:
            print("Calculated chemical potential limits: \n")
            print(df)

        return df

    @property
    def chem_limits(self) -> dict:
        """Returns the calculated chemical potential limits"""
        if not hasattr(self, "_chem_limits"):
            self.calculate_chempots()
        return self._chem_limits

    @property
    def intrinsic_chem_limits(self) -> dict:
        """Returns the calculated intrinsic chemical potential limits"""
        if not hasattr(self, "_intrinsic_chem_limits"):
            self.calculate_chempots()
        return self._intrinsic_chem_limits

    @property
    def intrinsic_phase_diagram(self) -> dict:
        """Returns the calculated intrinsic phase diagram"""
        if not hasattr(self, "_intrinsic_phase_diagram"):
            self.calculate_chempots()
        return self._intrinsic_phase_diagram

    def cplap_input(self, dependent_variable=None, filename="input.dat"):
        """For completeness' sake, automatically saves to input.dat for cplap
        Args:
            dependent_variable (str) Pick one of the variables as dependent, the first element is
                chosen from the composition if this isn't set
            filename (str): filename, should end in .dat
        Returns
            None, writes input.dat file
        """
        if not hasattr(self, "chem_limits"):
            self.calculate_chempots(verbose=False)

        with open(filename, "w") as f:
            with contextlib.redirect_stdout(f):
                # get lowest energy bulk phase
                bulk_entries = [
                    sub_dict
                    for sub_dict in self.data
                    if self.bulk_composition.reduced_composition
                    == Composition(sub_dict["formula"]).reduced_composition
                ]
                bulk_entry = min(bulk_entries, key=lambda x: x["formation_energy"])
                print(
                    f"{len(self.bulk_composition.as_dict())}  # number of elements in bulk"
                )
                for k, v in self.bulk_composition.as_dict().items():
                    print(int(v), k, end=" ")
                print(
                    f"{bulk_entry['formation_energy']}  # num_atoms, element, formation_energy "
                    "(bulk)"
                )

                if dependent_variable is not None:
                    print(f"{dependent_variable}  # dependent variable (element)")
                else:
                    print(f"{self.elemental[0]}  # dependent variable (element)")

                # get only the lowest energy entries of compositions in self.data which are on a
                # facet in self._intrinsic_chem_limits
                bordering_phases = {
                    phase
                    for facet in self._chem_limits["facets"]
                    for phase in facet.split("-")
                }
                entries_for_cplap = [
                    entry_dict
                    for entry_dict in self.data
                    if entry_dict["formula"] in bordering_phases
                    and Composition(entry_dict["formula"]).reduced_composition
                    != self.bulk_composition.reduced_composition
                ]
                # cull to only the lowest energy entries of each composition
                culled_cplap_entries = {}
                for entry in entries_for_cplap:
                    reduced_comp = Composition(entry["formula"]).reduced_composition
                    if reduced_comp not in culled_cplap_entries:
                        culled_cplap_entries[reduced_comp] = entry
                    else:
                        if (
                            entry["formation_energy"]
                            < culled_cplap_entries[reduced_comp]["formation_energy"]
                        ):
                            culled_cplap_entries[reduced_comp] = entry

                print(f"{len(culled_cplap_entries)}  # number of bordering phases")
                for i in culled_cplap_entries.values():
                    print(
                        f"{len(Composition(i['formula']).as_dict())}  # number of elements in "
                        "phase:"
                    )
                    for k, v in Composition(i["formula"]).as_dict().items():
                        print(int(v), k, end=" ")
                    print(
                        f"{i['formation_energy']}  # num_atoms, element, formation_energy"
                    )


def combine_extrinsic(first, second, extrinsic_species):
    # TODO: Can we just integrate this to `CompetingPhaseAnalyzer`, so you just pass in a list of
    # extrinsic species and it does the right thing?
    """
    Combines chemical limits for different extrinsic species using chemical limits json file from
    ChemicalPotentialAnalysis. Usage explained in the example jupyter notebook
    Args:
        first (dict): First chemical potential dictionary, it can contain extrinsic species other
        than the set extrinsic species
        second (dict): Second chemical potential dictionary, it must contain the extrinsic species
        extrinsic_species (str): Extrinsic species in the second dictionary
    Returns:
        dict
    """
    keys = ["elemental_refs", "facets", "facets_wrt_el_refs"]
    if not all(key in first for key in keys):
        raise KeyError(
            "the first dictionary doesn't contain the correct keys - it should include "
            "elemental_refs, facets and facets_wrt_el_refs"
        )

    if not all(key in second for key in keys):
        raise KeyError(
            "the second dictionary doesn't contain the correct keys - it should include "
            "elemental_refs, facets and facets_wrt_el_refs"
        )

    if extrinsic_species not in second["elemental_refs"].keys():
        raise ValueError("extrinsic species is not present in the second dictionary")

    cpa1 = copy.deepcopy(first)
    cpa2 = copy.deepcopy(second)
    new_facets = {}
    for (k1, v1), (k2, v2) in zip(
        list(cpa1["facets"].items()), list(cpa2["facets"].items())
    ):
        if k2.rsplit("-", 1)[0] in k1:
            new_key = k1 + "-" + k2.rsplit("-", 1)[1]
        else:
            raise ValueError(
                "The facets aren't matching, make sure you've used the correct dicitonary"
            )

        v1[extrinsic_species] = v2.pop(extrinsic_species)
        new_facets[new_key] = v1

    new_facets_wrt_elt = {}
    for (k1, v1), (k2, v2) in zip(
        list(cpa1["facets_wrt_el_refs"].items()),
        list(cpa2["facets_wrt_el_refs"].items()),
    ):
        if k2.rsplit("-", 1)[0] in k1:
            new_key = k1 + "-" + k2.rsplit("-", 1)[1]
        else:
            raise ValueError(
                "The facets aren't matching, make sure you've used the correct dicitonary"
            )

        v1[extrinsic_species] = v2.pop(extrinsic_species)
        new_facets_wrt_elt[new_key] = v1

    new_elements = copy.deepcopy(cpa1["elemental_refs"])
    new_elements[extrinsic_species] = copy.deepcopy(cpa2["elemental_refs"])[
        extrinsic_species
    ]

    new_dict = {}
    new_dict = {
        "elemental_refs": new_elements,
        "facets": new_facets,
        "facets_wrt_el_refs": new_facets_wrt_elt,
    }

    return new_dict



#!/usr/bin/env python


import os
import warnings
from collections import defaultdict
from itertools import combinations
from math import exp, pi, sqrt

import numpy as np
from pymatgen.core import Element
from pymatgen.core.structure import PeriodicSite, Structure
from pymatgen.entries.computed_entries import ComputedStructureEntry
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from doped.pycdt.corrections.finite_size_charge_correction import (
    get_correction_freysoldt,
    get_correction_kumagai,
)
from doped.pycdt.utils.parse_calculations import SingleDefectParser
from doped.pycdt.utils.units import conv, hbar, kb

warnings.simplefilter("default")
warnings.filterwarnings("ignore", message="`np.int` is a deprecated alias for the builtin `int`")
warnings.filterwarnings("ignore", message="Use get_magnetic_symmetry()")


def freysoldt_correction_from_paths(
    defect_file_path, bulk_file_path, dielectric, defect_charge, plot=False
):
    """
    A function for performing the Freysoldt correction with a set of file paths.
    If this correction is used, please reference Freysoldt's original paper.
    doi: 10.1103/PhysRevLett.102.016402

    Does not require transformation.json file to exist in file path.

    :param defect_file_path (str): file path to defect folder of interest
    :param bulk_file_path (str): file path to bulk folder of interest
    :param dielectric (float or 3x3 matrix): Dielectric constant (or tensor) for the structure
    :param defect_charge (int): charge of defect structure of interest
    :param plot (bool): allow for plotting electrostatic potential
    :return:
        Dictionary of Freysoldt Correction for defect
    """
    sdp = SingleDefectParser.from_paths(
        defect_file_path, bulk_file_path, dielectric, defect_charge
    )
    _ = sdp.freysoldt_loader()
    if plot:
        print(f"{sdp.defect_entry.name}, charge = {defect_charge}")
    correction = get_correction_freysoldt(sdp.defect_entry, dielectric, plot=plot)

    return correction


def kumagai_correction_from_paths(
    defect_file_path, bulk_file_path, dielectric, defect_charge, plot=False
):
    """
    A function for performing the Kumagai correction with a set of file paths.
    If this correction is used, please reference Kumagai and Oba's original paper
    (doi: 10.1103/PhysRevB.89.195205) as well as Freysoldt's original
    paper (doi: 10.1103/PhysRevLett.102.016402

    Does not require transformation.json file to exist in file path.

    :param defect_file_path (str): file path to defect folder of interest
    :param bulk_file_path (str): file path to bulk folder of interest
    :param dielectric (float or 3x3 matrix): Dielectric constant (or tensor) for the structure
    :param defect_charge (int): charge of defect structure of interest
    :param plot (bool): allow for plotting electrostatic potential
    :return:
        Dictionary of Kumagai Correction for defect
    """
    sdp = SingleDefectParser.from_paths(
        defect_file_path, bulk_file_path, dielectric, defect_charge
    )
    _ = sdp.kumagai_loader()
    plt_title = (
        os.path.join(
            defect_file_path, "{}_chg_{}".format(sdp.defect_entry.name, defect_charge)
        )
        if plot
        else None
    )
    correction = get_correction_kumagai(sdp.defect_entry, dielectric, title=plt_title)

    return correction


class ComputedDefect(object):
    """
    Holds all the info concerning a defect computation:
    composition+structure, energy, correction on energy and name
    """

    def __init__(
        self,
        entry_defect,
        site_in_bulk,
        multiplicity=None,
        supercell_size=(1, 1, 1),
        charge=0.0,
        charge_correction=0.0,
        other_correction=0.0,
        name=None,
    ):
        """
        Args:
            entry_defect:
                An ComputedStructureEntry object corresponding to the
                defect supercell
            site_in_bulk:
                Site of the defect in bulk supercell. Defect positions
                are often required to perform posteriori corrections
            multiplicity:
                Multiplicity of defect site in a cell. Useful to
                evaluate defect concentrations
            supercell_size:
                Size of the defect supercell in terms of unit cell
            charge:
                The charge of the defect
            charge_correction:
                Correction to the energy due to charge
            other_correction:
                Correction to the energy due to other factors
            name:
                The name of the defect
        """

        self.entry = entry_defect
        self.site = site_in_bulk
        self.multiplicity = multiplicity
        self.supercell_size = supercell_size
        self.charge = charge
        self.charge_correction = charge_correction  # Can be added after initialization
        self.other_correction = other_correction
        self.name = name
        if self.name:
            self.full_name = self.name + "_" + str(charge)
        else:
            self.full_name = "defect_" + str(charge)
        warnings.warn(
            "Replaced PyCDT usage of ComputedDefect objects with "
            "DefectEntry objects from pymatgen.analysis.defects.core\n"
            "Will remove ComputedDefect with Version 2.5 of PyCDT.",
            DeprecationWarning,
        )

    def as_dict(self):
        return {
            "entry": self.entry.as_dict(),
            "site": self.site.as_dict(),
            "multiplicity": self.multiplicity,
            "supercell_size": self.supercell_size,
            "charge": self.charge,
            "charge_correction": self.charge_correction,
            "other_correction": self.other_correction,
            "name": self.name,
            "full_name": self.full_name,
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(
            ComputedStructureEntry.from_dict(d["entry"]),
            PeriodicSite.from_dict(d["site"]),
            multiplicity=d.get("multiplicity", None),
            supercell_size=d.get("supercell_size", [1, 1, 1]),
            charge=d.get("charge", 0.0),
            charge_correction=d.get("charge_correction", 0.0),
            other_correction=d.get("other_correction", 0.0),
            name=d.get("name", None),
        )


class DefectsAnalyzer(object):
    """
    a class aimed at performing standard analysis of defects
    """

    def __init__(self, entry_bulk, e_vbm, mu_elts, band_gap):
        """
        Args:
            entry_bulk:
                the bulk data as an Entry
            e_vbm:
                the energy of the vbm (in eV)
            mu_elts:
                a dictionnary of {Element:value} giving the chemical
                potential of each element
            band_gap:
                the band gap (in eV)
        """
        self._entry_bulk = entry_bulk
        self._e_vbm = e_vbm
        self._mu_elts = mu_elts
        self._band_gap = band_gap
        self._defects = []
        self._formation_energies = []
        warnings.warn(
            "Replaced PyCDT usage of DefectsAnalyzer objects with "
            "DefectPhaseDiagram objects from pymatgen.analysis.defects.thermodynamics\n"
            "Will remove DefectsAnalyzer with Version 2.5 of PyCDT.",
            DeprecationWarning,
        )

    def as_dict(self):
        d = {
            "entry_bulk": self._entry_bulk.as_dict(),
            "e_vbm": self._e_vbm,
            "mu_elts": {k.symbol: v for k, v in self._mu_elts.items()},
            "band_gap": self._band_gap,
            "defects": [d.as_dict() for d in self._defects],
            "formation_energies": self._formation_energies,
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
        }
        return d

    @classmethod
    def from_dict(cls, d):
        struct = d["entry_bulk"]["structure"]
        struct = (
            struct if isinstance(struct, Structure) else Structure.from_dict(struct)
        )
        entry_bulk = ComputedStructureEntry(struct, d["entry_bulk"]["energy"])
        analyzer = DefectsAnalyzer(
            entry_bulk,
            d["e_vbm"],
            {Element(el): d["mu_elts"][el] for el in d["mu_elts"]},
            d["band_gap"],
        )
        for ddict in d["defects"]:
            analyzer.add_computed_defect(ComputedDefect.from_dict(ddict))
        return analyzer

    def add_computed_defect(self, defect):
        """
        add a parsed defect to the analyzer
        Args:
            defect:
                a ComputedDefect object
        """
        self._defects.append(defect)
        self._compute_form_en()

    def change_charge_correction(self, i, correction):
        """
        Change the charge correction for defect at index i
        Args:
            i:
                Index of defects
            correction:
                New correction to be applied for defect
        """
        self._defects[i].charge_correction = correction
        self._compute_form_en()

    def change_other_correction(self, i, correction):
        """
        Change the charge correction for defect at index i
        Args:
            i:
                Index of defects
            correction:
                New correction to be applied for defect
        """
        self._defects[i].other_correction = correction
        self._compute_form_en()

    def _get_all_defect_types(self):
        to_return = []
        for d in self._defects:
            if d.name not in to_return:
                to_return.append(d.name)
        return to_return

    def _compute_form_en(self):
        """
        compute the formation energies for all defects in the analyzer
        """
        self._formation_energies = []
        for d in self._defects:
            # compensate each element in defect with the chemical potential
            mu_needed_coeffs = {}
            for elt in d.entry.composition.elements:
                el_def_comp = d.entry.composition[elt]
                el_blk_comp = self._entry_bulk.composition[elt]
                mu_needed_coeffs[Element(elt)] = el_blk_comp - el_def_comp

            sum_mus = 0.0
            for elt in mu_needed_coeffs:
                sum_mus += mu_needed_coeffs[elt] * self._mu_elts[elt]

            self._formation_energies.append(
                d.entry.energy
                - self._entry_bulk.energy
                + sum_mus
                + d.charge * self._e_vbm
                + d.charge_correction
                + d.other_correction
            )

    def correct_bg_simple(self, vbm_correct, cbm_correct):
        """
        correct the band gap in the analyzer.
        We assume the defects level remain the same when moving the
        band edges
        Args:
            vbm_correct:
                The correction on the vbm as a positive number. e.g.,
                if the VBM goes 0.1 eV down vbm_correct=0.1
            cbm_correct:
                The correction on the cbm as a positive number. e.g.,
                if the CBM goes 0.1 eV up cbm_correct=0.1

        """
        self._band_gap = self._band_gap + cbm_correct + vbm_correct
        self._e_vbm = self._e_vbm - vbm_correct
        self._compute_form_en()

    def get_transition_levels(self):
        """
        Charge transition levels are computed
        :return: Transition levels for each pair of the defects.
        If any pair is missing, the transition level for that pair is
        not within the band gap.
        """
        xlim = (-0.5, self._band_gap + 1.5)
        nb_steps = 1000
        x = np.arange(xlim[0], xlim[1], (xlim[1] - xlim[0]) / nb_steps)

        y = defaultdict(defaultdict)
        for i, dfct in enumerate(self._defects):
            yval = self._formation_energies[i] + dfct.charge * x
            y[dfct.name][dfct.charge] = yval

        transit_levels = defaultdict(defaultdict)
        for dfct_name in y:
            q_ys = y[dfct_name]
            for qpair in combinations(q_ys.keys(), 2):
                qpair_s = tuple(sorted(list(qpair)))
                y_absdiff = abs(q_ys[qpair_s[1]] - q_ys[qpair_s[0]])
                if y_absdiff.min() < 0.4:
                    transit_levels[dfct_name][qpair_s] = x[np.argmin(y_absdiff)]
        return transit_levels

    def _get_form_energy(self, ef, i):
        return self._formation_energies[i] + self._defects[i].charge * ef

    def get_formation_energies(self, ef=0.0):
        """
        Get the defect formation energies for a given Fermi level
        Args:
            ef:
                the fermi level in eV (with respect to the VBM)
        Returns:
            a list of dict of {'name': defect name, 'charge': defect charge
                               'energy': defect formation energy in eV}
        """
        energies = []
        i = 0
        for i, d in enumerate(self._defects):
            energies.append(
                {
                    "name": d.name,
                    "charge": d.charge,
                    "energy": self._get_form_energy(ef, i),
                }
            )
        return energies

    def get_defects_concentration(self, temp=300, ef=0.0):
        """
        Get the defect concentration for a temperature and Fermi level.
        Note: This method has an approximation and can fail
        [Ref: PRB 86, 144109 (2012)]
        Args:
            temp:
                the temperature in K
            Ef:
                the fermi level in eV (with respect to the VBM)
        Returns:
            A list of dict of {'name': defect name, 'charge': defect charge
                               'conc': defects concentration in m-3}
        """
        conc = []
        struct = self._entry_bulk.structure
        for i, d in enumerate(self._defects):
            cell_multiplier = np.prod(d.supercell_size)
            n = d.multiplicity * cell_multiplier * 1e30 / struct.volume
            conc.append(
                {
                    "name": d.name,
                    "charge": d.charge,
                    "conc": n * exp(-self._get_form_energy(ef, i) / (kb * temp)),
                }
            )

        return conc

    def get_defects_concentration_old(self, temp=300, ef=0.0):
        """
        get the defect concentration for a temperature and Fermi level
        Written when the site multiplicity is not supplied with ComputedDefect
        Args:
            temp:
                the temperature in K
            Ef:
                the fermi level in eV (with respect to the VBM)
        Returns:
            a list of dict of {'name': defect name, 'charge': defect charge
                               'conc': defects concentration in m-3}
        """
        conc = []
        spga = SpacegroupAnalyzer(self._entry_bulk.structure, symprec=1e-1)
        struct = spga.get_symmetrized_structure()
        i = 0
        for d in self._defects:
            df_coords = d.site.frac_coords
            target_site = None
            for s in struct.sites:
                sf_coords = s.frac_coords
                if (
                    abs(s.frac_coords[0] - df_coords[0]) < 0.1
                    and abs(s.frac_coords[1] - df_coords[1]) < 0.1
                    and abs(s.frac_coords[2] - df_coords[2]) < 0.1
                ):
                    target_site = s
                    break
            equiv_site_no = len(struct.find_equivalent_sites(target_site))
            n = equiv_site_no * 1e30 / struct.volume
            conc.append(
                {
                    "name": d.name,
                    "charge": d.charge,
                    "conc": n * exp(-self._get_form_energy(ef, i) / (kb * temp)),
                }
            )
            i += 1
        return conc

    def _get_dos(self, e, m1, m2, m3, e_ext):
        return sqrt(2) / (pi**2 * hbar**3) * sqrt(m1 * m2 * m3) * sqrt(e - e_ext)

    def _get_dos_fd_elec(self, e, ef, t, m1, m2, m3):
        return (
            conv
            * (2.0 / (exp((e - ef) / (kb * t)) + 1))
            * (sqrt(2) / (pi**2))
            * sqrt(m1 * m2 * m3)
            * sqrt(e - self._band_gap)
        )

    def _get_dos_fd_hole(self, e, ef, t, m1, m2, m3):
        return (
            conv
            * (exp((e - ef) / (kb * t)) / (exp((e - ef) / (kb * t)) + 1))
            * (2.0 * sqrt(2) / (pi**2))
            * sqrt(m1 * m2 * m3)
            * sqrt(-e)
        )

    def _get_qd(self, ef, t):
        summation = 0.0
        for d in self.get_defects_concentration(t, ef):
            summation += d["charge"] * d["conc"]
        return summation

    def get_qi(self, ef, t, m_elec, m_hole):
        from scipy import integrate as intgrl

        elec_den_fn = lambda e: self._get_dos_fd_elec(
            e, ef, t, m_elec[0], m_elec[1], m_elec[2]
        )
        hole_den_fn = lambda e: self._get_dos_fd_hole(
            e, ef, t, m_hole[0], m_hole[1], m_hole[2]
        )

        bg = self._band_gap
        elec_count = -intgrl.quad(elec_den_fn, bg, bg + 5)[0]
        hole_count = intgrl.quad(hole_den_fn, -5, 0.0)[0]

        return elec_count + hole_count

    def _get_qtot(self, ef, t, m_elec, m_hole):
        return self._get_qd(ef, t) + self.get_qi(ef, t, m_elec, m_hole)

    def get_eq_ef(self, t, m_elec, m_hole):
        """
        access to equilibrium values of Fermi level and concentrations
        in defects and carriers obtained by self-consistent solution of
        charge balance + defect and carriers concentrations
        Args:
            t: temperature in K
            m_elec: electron effective mass as a 3 value list
                    (3 eigenvalues for the tensor)
            m_hole:: hole effective mass as a 3 value list
                    (3 eigenvalues for the tensor)
        Returns:
            a dict with {
                'ef':eq fermi level,
                'Qi': the concentration of carriers
                      (positive for holes, negative for e-) in m^-3,
                'conc': the concentration of defects as a list of dicts
                }
        """
        from scipy.optimize import bisect

        e_vbm = self._e_vbm
        e_cbm = self._e_vbm + self._band_gap
        ef = bisect(lambda e: self._get_qtot(e, t, m_elec, m_hole), 0, self._band_gap)
        return {
            "ef": ef,
            "Qi": self.get_qi(ef, t, m_elec, m_hole),
            "QD": self._get_qd(ef, t),
            "conc": self.get_defects_concentration(t, ef),
        }

    def get_non_eq_ef(self, tsyn, teq, m_elec, m_hole):
        """
        access to the non-equilibrium values of Fermi level and
        concentrations in defects and carriers obtained by
        self-consistent solution of charge balance + defect and
        carriers concentrations

        Implemented following Sun, R., Chan, M. K. Y., Kang, S.,
        and Ceder, G. (2011). doi:10.1103/PhysRevB.84.035212

        Args:
            tsyn: the synthesis temperature in K
            teq: the temperature of use in K
            m_elec: electron effective mass as a 3 value list
                    (3 eigenvalues for the tensor)
            m_hole: hole effective mass as a 3 value list
                    (3 eigenvalues for the tensor)
        Returns:
            a dict with {
                'ef':eq fermi level,
                'Qi': the concentration of carriers
                      (positive for holes, negative for e-) in m^-3,
                'conc': the concentration of defects as a list of dict
                }
        """
        from scipy.optimize import bisect

        eqsyn = self.get_eq_ef(tsyn, m_elec, m_hole)
        cd = {}
        for c in eqsyn["conc"]:
            if c["name"] in cd:
                cd[c["name"]] += c["conc"]
            else:
                cd[c["name"]] = c["conc"]
        ef = bisect(
            lambda e: self._get_non_eq_qtot(cd, e, teq, m_elec, m_hole),
            -1.0,
            self._band_gap + 1.0,
        )
        return {
            "ef": ef,
            "Qi": self.get_qi(ef, teq, m_elec, m_hole),
            "conc_syn": eqsyn["conc"],
            "conc": self._get_non_eq_conc(cd, ef, teq),
        }

    def _get_non_eq_qd(self, cd, ef, t):
        sum_tot = 0.0
        for n in cd:
            sum_d = 0.0
            sum_q = 0.0
            i = 0
            for d in self._defects:
                if d.name == n:
                    sum_d += exp(-self._get_form_energy(ef, i) / (kb * t))
                    sum_q += d.charge * exp(-self._get_form_energy(ef, i) / (kb * t))
                i += 1
            sum_tot += cd[n] * sum_q / sum_d
        return sum_tot

    def _get_non_eq_conc(self, cd, ef, t):
        sum_tot = 0.0
        res = []
        for n in cd:
            sum_tot = 0
            i = 0
            for d in self._defects:
                if d.name == n:
                    sum_tot += exp(-self._get_form_energy(ef, i) / (kb * t))
                i += 1
            i = 0
            for d in self._defects:
                if d.name == n:
                    res.append(
                        {
                            "name": d.name,
                            "charge": d.charge,
                            "conc": cd[n]
                            * exp(-self._get_form_energy(ef, i) / (kb * t))
                            / sum_tot,
                        }
                    )
                i += 1
        return res

    def _get_non_eq_qtot(self, cd, ef, t, m_elec, m_hole):
        return self._get_non_eq_qd(cd, ef, t) + self.get_qi(ef, t, m_elec, m_hole)

    def correct_bg(self, dict_levels, vbm_correct, cbm_correct):
        """
        NOTE from developers: This code uses deprecated concepts and
            will not be used or maintained going forward (as of 12/15/2017).
            However, we are keeping function here to allow for
            current users to make use of it...

        correct the band gap in the analyzer and make sure the levels move
        accordingly.
        There are two types of defects vbm_like and cbm_like and we need
        to provide a formal oxidation state
        The vbm-like will follow the vbm and the cbm_like the cbm. If nothing
        is specified the defect transition level does not move
        Args:
            dict_levels: a dictionary of type {defect_name:
            {'type':type_of_defect,'q*':formal_ox_state}}
            Where type_of_defect is a string: 'vbm_like' or 'cbm_like'
        """
        self._band_gap = self._band_gap + cbm_correct + vbm_correct
        self._e_vbm = self._e_vbm - vbm_correct
        self._compute_form_en()
        for i in range(len(self._defects)):
            name = self._defects[i].name
            if not name in dict_levels:
                continue

            if dict_levels[name]["type"] == "vbm_like":
                z = self._defects[i].charge - dict_levels[name]["q*"]
                self._formation_energies[i] += z * vbm_correct
            if dict_levels[name]["type"] == "cbm_like":
                z = dict_levels[name]["q*"] - self._defects[i].charge
                self._formation_energies[i] += z * cbm_correct

    def get_defect_occupancies(self):
        """
        NOTE from developers: This code uses deprecated concepts and
            will not be used or maintained going forward (as of 12/15/2017).
            However, we are keeping function here to allow for
            current users to make use of it...

        Defect occupancies with respect to defect chargest.
        The assumption is that the highest charge of defect numerically
        has zero occupancy:
        Ex: In Cr2O3, V_O has 0 occupancy for +2 q and 2 occupancy for 0 q.
            V_{Cr} has 0 occupancy for 0 q and 3 occupancy for -3 q
        Caution: Didn't check for semiconductor with large # of defect q's.
        Returns:
            Defect occupancies as a nested dict
        """
        charges = defaultdict(list)
        for dfct in self._defects:
            charges[dfct.name].append(dfct.charge)

        occupancies = defaultdict(lambda: defaultdict(int))
        for dfct_name in charges:
            for i, q in enumerate(sorted(charges[dfct_name], reverse=True)):
                occupancies[dfct_name][q] = i
            occupancies[dfct_name]["0_occupancy"] = sorted(
                charges[dfct_name], reverse=True
            )[0]
        return occupancies
